import glob
import re
from pathlib import Path
from shutil import rmtree
from typing import cast

import pytest

import icechunk as ic
import zarr


def test_error_message_when_snapshot_deleted(tmpdir: Path) -> None:
    tmpdir = Path(tmpdir)
    storage = ic.local_filesystem_storage(str(tmpdir))
    repo = ic.Repository.create(storage=storage)

    rmtree(tmpdir / "snapshots")

    repo = ic.Repository.open(storage=storage)
    # we check error includes the spans for ancestry and fetch_snapshot
    with pytest.raises(
        ic.IcechunkError, match=re.compile("fetch_snapshot.*ancestry", re.DOTALL)
    ):
        repo.ancestry(branch="main")


def test_error_message_when_manifest_file_altered(tmpdir: Path) -> None:
    tmpdir = Path(tmpdir)
    storage = ic.local_filesystem_storage(str(tmpdir))
    repo = ic.Repository.create(storage=storage)

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    array = group.create_array("array", shape=(100, 100), chunks=(10, 10), dtype="i4")

    array[:, :] = 42
    session.commit("commit 1")

    manifest_path = glob.glob(f"{tmpdir}/manifests/*")[0]
    with open(manifest_path, "w") as file:
        file.write("invalid msgpack")

    repo = ic.Repository.open(storage=storage)

    session = repo.readonly_session(branch="main")
    store = session.store
    group = zarr.Group.open(store=store)
    fetched_array = cast(zarr.Array, group["array"])

    ## we check error includes the spans for ancestry and fetch_snapshot
    with pytest.raises(
        ic.IcechunkError, match=re.compile("fetch_manifest.*get", re.DOTALL)
    ):
        fetched_array[0]



#!/usr/bin/env python3

import datetime
import itertools
import json
import operator
import textwrap
from dataclasses import dataclass
from functools import partial
from typing import Any, Self

import numpy as np
import pytest

from zarr.core.buffer import Buffer, default_buffer_prototype

pytest.importorskip("hypothesis")
pytest.importorskip("xarray")

import copy

import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import pytest
from hypothesis import assume, note
from hypothesis.stateful import (
    Bundle,
    RuleBasedStateMachine,
    Settings,
    consumes,
    initialize,
    invariant,
    precondition,
    rule,
)

import zarr.testing.strategies as zrst
from icechunk import IcechunkError, Repository, SnapshotInfo, in_memory_storage
from zarr.testing.stateful import SyncStoreWrapper

# JSON file contents, keep it simple
simple_text = st.text(zrst.zarr_key_chars, min_size=1, max_size=5)
simple_attrs = st.dictionaries(
    simple_text,
    st.integers(min_value=-10_000, max_value=10_000),
    max_size=5,
)

DEFAULT_BRANCH = "main"


#########
# TODO: port to Zarr


class NewSyncStoreWrapper(SyncStoreWrapper):
    def list_prefix(self, prefix: str) -> list[str]:
        return self._sync_iter(self.store.list_prefix(prefix))


#####

MAX_TEXT_SIZE = 120

keys = st.lists(zrst.node_names, min_size=1, max_size=4).map("/".join)
metadata_paths = keys.map(lambda x: x + "/zarr.json")
v3_array_metadata = zrst.array_metadata(
    zarr_formats=st.just(3),
    array_shapes=partial(
        npst.array_shapes, max_dims=4, min_side=1, max_side=5
    ),  # set min_side to one
    attributes=simple_attrs,
).map(lambda x: x.to_buffer_dict(prototype=default_buffer_prototype())["zarr.json"])


@dataclass
class ExpireInfo:
    expired_snapshots: set[str]
    deleted_branches: set[str]
    deleted_tags: set[str]


@dataclass
class CommitModel:
    id: str
    written_at: datetime.datetime
    store: dict[str, Any]
    parent_id: str

    @classmethod
    def from_snapshot_and_store(cls, snap: SnapshotInfo, store: dict[str, Any]) -> Self:
        return cls(
            id=snap.id, written_at=snap.written_at, parent_id=snap.parent_id, store=store
        )


@dataclass
class TagModel:
    commit_id: str
    # message: str | None


class Model:
    def __init__(self, **kwargs) -> None:
        self.store: dict[str, Any] = {}  #

        self.initial_snapshot_id: str | None = None
        self.changes_made: bool = False

        self.HEAD: None | str = None
        self.branch: None | str = None

        self.commits: dict[str, CommitModel] = {}
        self.tags: dict[str, TagModel] = {}
        # TODO: This is only tracking the HEAD,
        # Should we model the branch as an ordered list of commits?
        self.branches: dict[str, str] = {}

        # a tag once created, can never be recreated even after expiration
        self.created_tags: set[str] = set()

    def __repr__(self) -> str:
        return textwrap.dedent(f"""
        <Model branch: {self.branch!r}, HEAD: {self.HEAD!r}, changes_made: {self.changes_made!r}>
        Branches: {tuple(self.branches.keys())!r}
        Tags: {tuple(self.tags.keys())!r}""").strip("\n")

    def __setitem__(self, key: str, value: Buffer) -> None:
        # Icechunk doesn't overwrite docs with the same value
        # and we need to keep `changes_made` in sync.
        # Icechunk checks after decoding the metadata to rust Structs so we must do the same.
        # Different byte strings can decode to the same json dict (order of user attributes may be different)
        if key in self.store and json.loads(self.store[key].to_bytes()) == json.loads(
            value.to_bytes()
        ):
            note(f"skipping setting {key!r}, value is unchanged")
            return
        self.changes_made = True
        self.store[key] = value

    def __getitem__(self, key: str) -> Buffer:
        return self.store[key]

    @property
    def has_commits(self) -> bool:
        return bool(self.commits)

    def commit(self, snap: SnapshotInfo) -> None:
        ref = snap.id
        self.commits[ref] = CommitModel.from_snapshot_and_store(
            snap, copy.deepcopy(self.store)
        )
        self.changes_made = False
        self.HEAD = ref

        assert self.branch is not None
        self.branches[self.branch] = ref

    def checkout_commit(self, ref: str) -> None:
        assert str(ref) in self.commits
        # deepcopy so that we allow changes, but the committed store remains unchanged
        # TODO: consider Frozen stores in self.commit?
        self.store = copy.deepcopy(self.commits[ref].store)
        self.changes_made = False
        self.HEAD = ref
        self.branch = None

    def create_branch(self, name: str, commit: str) -> None:
        assert commit in self.commits
        self.branches[name] = commit

    def checkout_branch(self, ref: str) -> None:
        self.checkout_commit(self.branches[ref])
        self.branch = ref

    def reset_branch(self, branch: str, commit: str) -> None:
        assert commit in self.commits
        self.branches[branch] = commit

    def delete_branch(self, branch_name: str) -> None:
        del self.branches[branch_name]

    def delete_tag(self, tag: str) -> None:
        del self.tags[tag]

    def create_tag(self, tag_name: str, commit_id: str) -> None:
        assert commit_id in self.commits
        self.tags[tag_name] = TagModel(commit_id=str(commit_id))
        self.created_tags.add(tag_name)

    def checkout_tag(self, ref: str) -> None:
        self.checkout_commit(self.tags[str(ref)].commit_id)

    def list_prefix(self, prefix: str) -> tuple[str, ...]:
        assert prefix == ""
        return tuple(self.store)

    def refs_iter(self):
        tag_iter = map(operator.attrgetter("commit_id"), self.tags.values())
        return itertools.chain(self.branches.values(), tag_iter)

    def expire_snapshots(
        self,
        older_than: datetime.datetime,
        *,
        delete_expired_branches: bool,
        delete_expired_tags: bool,
    ) -> ExpireInfo:
        # model this exactly like icechunk does.
        # Start with named refs, and walk the ancestry back to the root.
        expired_snaps = set()
        edited_snaps = set()
        for commit_id in self.refs_iter():
            snap = self.commits[commit_id]
            if snap.parent_id == self.initial_snapshot_id:
                continue

            editable_snap = snap.id
            while commit_id != self.initial_snapshot_id:
                snap = self.commits[commit_id]
                if snap.written_at < older_than:
                    expired_snaps.add(snap.id)
                else:
                    editable_snap = snap.id
                commit_id = snap.parent_id
            if editable_snap != self.initial_snapshot_id:
                edited_snaps.add(editable_snap)
                self.commits[editable_snap].parent_id = self.initial_snapshot_id

        note(f"model {expired_snaps=!r}")

        if delete_expired_tags:
            tags_to_delete = {
                k
                for k, v in self.tags.items()
                if self.commits[v.commit_id].written_at < older_than
            }
            note(f"deleting tags {tags_to_delete=!r}")
            for tag in tags_to_delete:
                self.delete_tag(tag)
        else:
            tags_to_delete = set()

        if delete_expired_branches:
            branches_to_delete = {
                k
                for k, v in self.branches.items()
                if k != DEFAULT_BRANCH and self.commits[v].written_at < older_than
            }
            note(f"deleting branches {branches_to_delete=!r}")
            for branch in branches_to_delete:
                note(f"deleting {branch=!r}, {self.branches[branch]=!r}")
                self.delete_branch(branch)
        else:
            branches_to_delete = set()

        return ExpireInfo(
            expired_snapshots=expired_snaps,
            deleted_branches=branches_to_delete,
            deleted_tags=tags_to_delete,
        )

    def garbage_collect(self, older_than: datetime.datetime) -> set[str]:
        reachable_snaps = set((self.initial_snapshot_id,))
        for commit_id in self.refs_iter():
            while commit_id != self.initial_snapshot_id:
                reachable_snaps.add(commit_id)
                commit_id = self.commits[commit_id].parent_id

        deleted = set()
        for k in set(self.commits) - reachable_snaps:
            if self.commits[k].written_at < older_than:
                del self.commits[k]
                deleted.add(k)
        note(f"Deleted snapshots: {deleted!r}")
        return deleted


class VersionControlStateMachine(RuleBasedStateMachine):
    """
    We use bundles to track the state, since Hypothesis will then
    preferably draw the same value for different rules.
    e.g. create branch 'X', then delete branch 'X'
    """

    commits = Bundle("commits")
    tags = Bundle("tags")
    branches = Bundle("branches")

    def __init__(self) -> None:
        super().__init__()

        note("----------")
        self.model = Model()
        self.commit_times: list[datetime.datetime] = []

    @initialize(data=st.data(), target=branches)
    def initialize(self, data: st.DataObject) -> str:
        self.repo = Repository.create(in_memory_storage())
        self.session = self.repo.writable_session(DEFAULT_BRANCH)

        snap = next(iter(self.repo.ancestry(branch=DEFAULT_BRANCH)))
        note(f"Initial commit is {snap!r}")
        self.initial_snapshot = snap
        self.model.initial_snapshot_id = snap.id

        HEAD = self.repo.lookup_branch(DEFAULT_BRANCH)
        self.model.commits[HEAD] = CommitModel.from_snapshot_and_store(snap, {})
        self.model.HEAD = HEAD
        self.model.create_branch(DEFAULT_BRANCH, HEAD)
        self.model.checkout_branch(DEFAULT_BRANCH)

        # initialize with some data always
        # TODO: always setting array metadata, since we cannot overwrite an existing group's zarr.json
        #       with an array's zarr.json
        # TODO: consider adding a deeper understanding of the zarr model rather than just setting docs?
        self.set_doc(path="zarr.json", value=data.draw(v3_array_metadata))

        return DEFAULT_BRANCH

    def new_store(self) -> None:
        self.session = self.repo.writable_session(DEFAULT_BRANCH)

    @property
    def sync_store(self) -> NewSyncStoreWrapper:
        return NewSyncStoreWrapper(self.session.store)

    @rule(path=metadata_paths, value=v3_array_metadata)
    def set_doc(self, path: str, value: Buffer) -> None:
        note(f"setting path {path!r} with {value.to_bytes()!r}")
        if self.model.branch is not None:
            self.sync_store.set(path, value)
            self.model[path] = value
        else:
            # not at branch head, modifications not possible.
            with pytest.raises(IcechunkError, match="read-only store"):
                self.sync_store.set(path, value)

    @rule(message=st.text(max_size=MAX_TEXT_SIZE), target=commits)
    @precondition(lambda self: self.model.changes_made)
    def commit(self, message: str) -> str:
        branch = self.session.branch
        assert branch is not None
        commit_id = self.session.commit(message)
        snapinfo = next(iter(self.repo.ancestry(branch=branch)))
        assert snapinfo.id == commit_id
        self.session = self.repo.writable_session(branch)
        note(f"Created commit: {snapinfo!r}")
        self.model.commit(snapinfo)
        self.commit_times.append(snapinfo.written_at)
        return commit_id

    @rule(ref=commits)
    def checkout_commit(self, ref: str) -> None:
        if ref not in self.model.commits:
            note(f"Checking out commit {ref}, expecting error")
            with pytest.raises(IcechunkError):
                self.repo.readonly_session(snapshot_id=ref)
        else:
            note(f"Checking out commit {ref}")
            self.session = self.repo.readonly_session(snapshot_id=ref)
            assert self.session.read_only
            self.model.checkout_commit(ref)
            self.check_commit(ref)

    @rule(ref=tags)
    def checkout_tag(self, ref: str) -> None:
        """
        Tags and branches are combined here since checkout magically works for both.
        This test is relying on the model tracking tags and branches accurately.
        """
        maybe_tag = self.model.tags.get(ref)
        if maybe_tag and maybe_tag.commit_id in self.model.commits:
            note(f"Checking out tag {ref!r}")
            self.session = self.repo.readonly_session(tag=ref)
            assert self.session.read_only
            self.model.checkout_tag(ref)
        else:
            note("Expecting error.")
            with pytest.raises(IcechunkError):
                self.repo.readonly_session(tag=ref)

    @rule(ref=branches)
    def checkout_branch(self, ref: str) -> None:
        # TODO: sometimes readonly?
        if self.model.branches.get(ref) in self.model.commits:
            note(f"Checking out branch {ref!r}")
            self.session = self.repo.writable_session(ref)
            assert not self.session.read_only
            self.model.checkout_branch(ref)
        else:
            with pytest.raises(IcechunkError):
                note(f"Expecting error when checking out branch {ref!r}")
                self.repo.writable_session(ref)

    @rule(name=simple_text, commit=commits, target=branches)
    def create_branch(self, name: str, commit: str) -> str:
        note(f"Creating branch {name!r}")

        # we can create a tag and branch with the same name
        if name not in self.model.branches and commit in self.model.commits:
            self.repo.create_branch(name, commit)
            self.model.create_branch(name, commit)
        else:
            note("Expecting error.")
            with pytest.raises(IcechunkError):
                self.repo.create_branch(name, commit)
        # returning this `name` to the Bundle is OK even if the branch was not created
        # This will test out checking out and deleting a branch that does not exist.
        return name

    @precondition(lambda self: self.model.has_commits)
    @rule(name=simple_text, commit_id=commits, target=tags)
    def create_tag(self, name: str, commit_id: str) -> str:
        note(f"Creating tag {name!r} for commit {commit_id!r}")

        if (
            name in self.model.created_tags
            or name in self.model.tags
            or commit_id not in self.model.commits
        ):
            note("Expecting error.")
            with pytest.raises(IcechunkError):
                self.repo.create_tag(name, commit_id)
        else:
            self.repo.create_tag(name, commit_id)
            self.model.create_tag(name, commit_id)
        # returning this `name` to the Bundle is OK even if the tag was not created
        # This will test out checking out and deleting a tag that does not exist.
        return name

    @rule()
    def discard_changes(self) -> None:
        self.session.discard_changes()
        if self.model.branch is not None:
            self.model.checkout_branch(self.model.branch)
        else:
            self.model.checkout_commit(self.session.snapshot_id)

    # if there are changes in a session tied to the same branch
    # then an attempt to commit from that session will raise a conflict
    # (as is expected)
    @precondition(lambda self: not self.model.changes_made)
    @rule(branch=branches, commit=commits)
    def reset_branch(self, branch: str, commit: str) -> None:
        if branch not in self.model.branches or commit not in self.model.commits:
            note(f"resetting branch {branch}, expecting error.")
            with pytest.raises(IcechunkError):
                self.repo.reset_branch(branch, commit)
        else:
            note(
                f"resetting branch {branch} from {self.model.branches[branch]} to {commit}"
            )
            self.repo.reset_branch(branch, commit)
            self.model.reset_branch(branch, commit)
            self.maybe_checkout_branch(branch, checkout_branch=branch)

    def maybe_checkout_branch(
        self, modified_branch: str, *, checkout_branch: str = DEFAULT_BRANCH
    ) -> None:
        # TODO: if we delete the current branch, we hold on to an invalid session
        #       this is confusing UX. any attempts to commit will create a conflict
        if self.model.branch == modified_branch:
            self.session = self.repo.writable_session(checkout_branch)
            self.model.checkout_branch(checkout_branch)

    @rule(branch=consumes(branches))
    def delete_branch(self, branch: str) -> None:
        note(f"Deleting branch {branch!r}")
        if branch in self.model.branches:
            if branch == DEFAULT_BRANCH:
                note("Expecting error.")
                with pytest.raises(
                    IcechunkError, match=f"{DEFAULT_BRANCH} branch cannot be deleted"
                ):
                    self.repo.delete_branch(branch)
            else:
                self.repo.delete_branch(branch)
                self.model.delete_branch(branch)
                self.maybe_checkout_branch(branch)

        else:
            note("Expecting error.")
            with pytest.raises(IcechunkError):
                self.repo.delete_branch(branch)

    @precondition(lambda self: bool(self.commit_times))
    @rule(
        data=st.data(),
        delta=st.timedeltas(
            min_value=datetime.timedelta(days=-1), max_value=datetime.timedelta(days=1)
        ),
        delete_expired_branches=st.booleans(),
        delete_expired_tags=st.booleans(),
    )
    def expire_snapshots(
        self,
        data: st.DataObject,
        delta: int,
        delete_expired_branches: bool,
        delete_expired_tags: bool,
    ) -> None:
        commit_time = data.draw(st.sampled_from(self.commit_times))
        older_than = commit_time + delta
        note(
            f"Expiring snapshots {older_than=!r}, ({commit_time=!r}, {delta=!r}), {delete_expired_branches=!r}, {delete_expired_tags=!r}"
        )
        actual = self.repo.expire_snapshots(
            older_than,
            delete_expired_branches=delete_expired_branches,
            delete_expired_tags=delete_expired_tags,
        )
        note(f"repo  expired snaps={actual!r}")
        expected = self.model.expire_snapshots(
            older_than,
            delete_expired_branches=delete_expired_branches,
            delete_expired_tags=delete_expired_tags,
        )
        assert self.initial_snapshot.id not in actual
        assert actual == expected.expired_snapshots, (actual, expected)

        for branch in expected.deleted_branches:
            self.maybe_checkout_branch(branch)

    @precondition(lambda self: bool(self.commit_times))
    @rule(
        data=st.data(),
        # we delete based on snapshot created_at time, not flushed_at time
        # so for delta we just past an integer number of seconds to handle the
        # (small) difference in tests
        delta=st.integers(min_value=-86400, max_value=86400).filter(lambda x: x != 0),
    )
    def garbage_collect(self, data: st.DataObject, delta: int) -> None:
        commit_time = data.draw(st.sampled_from(self.commit_times))
        older_than = commit_time + datetime.timedelta(seconds=delta)
        note(
            f"running garbage_collect for {older_than=!r}, ({commit_time=!r}, {delta=!r})"
        )
        summary = self.repo.garbage_collect(older_than)
        expected = self.model.garbage_collect(older_than)
        assert summary.snapshots_deleted == len(expected), (
            summary.snapshots_deleted,
            expected,
        )

        # deleted a checked out snapshot :?
        for snapshot in expected:
            if self.model.HEAD == snapshot:
                self.session = self.repo.writable_session(DEFAULT_BRANCH)
                self.model.checkout_branch(DEFAULT_BRANCH)

    def check_commit(self, commit) -> None:
        assume(commit in self.model.commits)
        note(f"Checking {commit=!r}")
        expected = self.model.commits[commit]
        actual = self.repo.lookup_snapshot(commit)
        assert actual.id == expected.id
        assert actual.parent_id == expected.parent_id
        # even after expiration, written_at is unmodified
        assert actual.written_at == expected.written_at

    @invariant()
    def check_list_prefix_from_root(self) -> None:
        model_list = self.model.list_prefix("")
        ice_list = self.sync_store.list_prefix("")

        assert sorted(model_list) == sorted(ice_list)

        for k in model_list:
            # need to load to dict to compare since ordering of entries might differ
            expected = json.loads(self.model[k].to_bytes())
            value = self.sync_store.get(k, default_buffer_prototype())
            assert value is not None, k
            actual = json.loads(value.to_bytes())
            actual_fv = actual.pop("fill_value")
            expected_fv = expected.pop("fill_value")
            if actual_fv != expected_fv:
                # TODO: is this right? we are losing accuracy in serialization
                np.testing.assert_allclose(actual_fv, expected_fv)
            assert actual == expected

    @invariant()
    def check_tags(self) -> None:
        expected_tags = self.model.tags
        actual_tags = {
            tag: TagModel(commit_id=self.repo.lookup_tag(tag))
            for tag in self.repo.list_tags()
        }
        assert actual_tags == expected_tags

    @invariant()
    def check_branches(self) -> None:
        repo_branches = {k: self.repo.lookup_branch(k) for k in self.repo.list_branches()}
        assert self.model.branches == repo_branches

    @invariant()
    def check_ancestry(self) -> None:
        for branch in self.repo.list_branches():
            ancestry = list(self.repo.ancestry(branch=branch))
            ancestry_set = set([snap.id for snap in ancestry])
            # snapshot timestamps are monotonically decreasing in ancestry
            assert all(
                a.written_at > b.written_at for a, b in itertools.pairwise(ancestry)
            )
            # ancestry must be unique
            assert len(ancestry_set) == len(ancestry)
            # the initial snapshot is in every possible branch
            # this is a python-only invariant
            assert ancestry[-1] == self.initial_snapshot


VersionControlStateMachine.TestCase.settings = Settings(deadline=None)
VersionControlTest = VersionControlStateMachine.TestCase



"""
This example uses Dask as a task orchestration framework
to write or update an array in an Icechunk repository.
To write an Xarray object with dask array use `icechunk.xarray.to_icechunk`

To understand all the available options run:
```
python ./examples/dask_write.py --help
python ./examples/dask_write.py create --help
python ./examples/dask_write.py update --help
python ./examples/dask_write.py verify --help
```

Example usage:

```
python ./examples/dask_write.py --url s3://my-bucket/my-icechunk-repo create --t-chunks 100000 --x-chunks 4 --y-chunks 4 --chunk-x-size 112 --chunk-y-size 112
python ./examples/dask_write.py --url s3://my-bucket/my-icechunk-repo update --t-from 0 --t-to 1500 --workers 16
python ./examples/dask_write.py --url s3://my-bucket/my-icechunk-repo verify --t-from 0 --t-to 1500 --workers 16
```

The work is split into three different commands.
* `create` initializes the repository and the array, without writing any chunks. For this example
   we chose a 3D array that simulates a dataset that needs backfilling across its time dimension.
* `update` can be called multiple times to write a number of "pancakes" to the array.
  It does so by distributing the work among Dask workers, in small tasks, one pancake per task.
  The example invocation above, will write 1,500 pancakes using 16 Dask workers.
* `verify` can read a part of the array and check that it contains the required data.

Dask can be used to read and write to Icechunk from multiple processes and machines, we just need to use a lower level
Dask API based, for example, in `map/gather`. This mechanism is what we show in this example.
"""

import argparse
from dataclasses import dataclass
from typing import Any, cast
from urllib.parse import urlparse

import numpy as np

import icechunk
import zarr
from dask.distributed import Client
from dask.distributed import print as dprint


@dataclass
class Task:
    """A task distributed to Dask workers"""

    session: (
        icechunk.Session
    )  # The worker will use this Icechunk session to read/write to the dataset
    time: (
        int  # The position in the coordinate dimension where the read/write should happen
    )
    seed: int  # An RNG seed used to generate or recreate random data for the array


def generate_task_array(task: Task, shape: tuple[int, ...]) -> np.typing.ArrayLike:
    """Generates a randm array with the given shape and using the seed in the Task"""
    np.random.seed(task.seed)
    return np.random.rand(*shape)


def execute_write_task(task: Task) -> icechunk.Session:
    """Execute task as a write task.

    This will read the time coordinade from `task` and write a "pancake" in that position,
    using random data. Random data is generated using the task seed.

    Returns the Icechunk session after the write is done.

    As you can see Icechunk session can be passed to remote workers, and returned from them.
    The reason to return the session is that we'll need all the remote session, when they are
    done, to be able to do a single, global commit to Icechunk.
    """

    session = task.session
    store = session.store

    group = zarr.group(store=store, overwrite=False)
    array = cast(zarr.Array, group["array"])
    dprint(f"Writing at t={task.time}")
    data = generate_task_array(task, array.shape[0:2])
    array[:, :, task.time] = data
    dprint(f"Writing at t={task.time} done")
    return session


def execute_read_task(task: Task) -> None:
    """Execute task as a read task.

    This will read the time coordinade from `task` and read a "pancake" in that position.
    Then it will assert the data is valid by re-generating the random data from the passed seed.

    As you can see Icechunk sessions can be passed to remote workers.
    """

    session = task.session
    store = session.store
    group = zarr.group(store=store, overwrite=False)
    array = cast(zarr.Array, group["array"])

    actual = array[:, :, task.time]
    expected = generate_task_array(task, array.shape[0:2])
    np.testing.assert_array_equal(actual, expected)
    dprint(f"t={task.time} verified")


def storage_config(args: argparse.Namespace) -> dict[str, Any]:
    """Return the Icechunk S3 configuration map"""
    bucket = args.url.netloc
    prefix = args.url.path[1:]
    return {
        "bucket": bucket,
        "prefix": prefix,
        "region": "us-east-1",
    }


def repository_config(args: argparse.Namespace) -> icechunk.RepositoryConfig:
    """Return the Icechunk repo configuration.

    We lower the default to make sure we write chunks and not inline them.
    """
    config = icechunk.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 1
    return config


def create(args: argparse.Namespace) -> None:
    """Execute the create subcommand.

    Creates an Icechunk repo, a root group and an array named "array"
    with the shape passed as arguments.

    Commits the Icechunk repository when done.
    """
    repo = icechunk.Repository.create(
        storage=icechunk.s3_storage(**storage_config(args)),
        config=repository_config(args),
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    shape = (
        args.x_chunks * args.chunk_x_size,
        args.y_chunks * args.chunk_y_size,
        args.t_chunks * 1,
    )
    chunk_shape = (args.chunk_x_size, args.chunk_y_size, 1)

    group.create_array(
        "array",
        shape=shape,
        chunks=chunk_shape,
        dtype="f8",
        fill_value=float("nan"),
    )
    first_snapshot = session.commit("array created")
    print(f"Array initialized, snapshot {first_snapshot}")


def update(args: argparse.Namespace) -> None:
    """Execute the update subcommand.

    Uses Dask to write chunks to the Icechunk repository. Currently Icechunk cannot
    use the Dask array API (see https://github.com/earth-mover/icechunk/issues/185) but we
    can still use a lower level API to do the writes:
    * We split the work into small `Task`s, one 'pancake' per task, at a given t coordinate.
    * We use Dask's `map` to ship the `Task` to a worker
    * The `Task` includes a copy of the Icechunk Session, so workers can do the writes
    * When workers are done, they send their Session back
    * When all workers are done (Dask's `gather`), we take all Sessions and do a distributed commit in Icechunk
    """

    repo = icechunk.Repository.open(
        storage=icechunk.s3_storage(**storage_config(args)),
        config=repository_config(args),
    )

    session = repo.writable_session("main")
    fork = session.fork()

    tasks = [
        Task(
            session=fork,
            time=time,
            seed=time,
        )
        for time in range(args.t_from, args.t_to, 1)
    ]

    client = Client(n_workers=args.workers, threads_per_worker=1)

    map_result = client.map(execute_write_task, tasks)
    worker_sessions = client.gather(map_result)

    print("Starting distributed commit")
    # we can use the current session as the commit coordinator, because it doesn't have any pending changes,
    # all changes come from the tasks, Icechunk doesn't care about where the changes come from, the only
    # important thing is to not count changes twice
    session.merge(*worker_sessions)
    commit_res = session.commit("distributed commit")
    assert commit_res
    print("Distributed commit done")


def verify(args: argparse.Namespace) -> None:
    """Execute the verify subcommand.

    Uses Dask to read and verify chunks from the Icechunk repository. Currently Icechunk cannot
    use the Dask array API (see https://github.com/earth-mover/icechunk/issues/185) but we
    can still use a lower level API to do the verification:
    * We split the work into small `Task`s, one 'pancake' per task, at a given t coordinate.
    * We use Dask's `map` to ship the `Task` to a worker
    * The `Task` includes a copy of the Icechunk Store, so workers can do the Icechunk reads
    """
    repo = icechunk.Repository.open(
        storage=icechunk.s3_storage(**storage_config(args)),
        config=repository_config(args),
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=False)
    array = cast(zarr.Array, group["array"])
    print(f"Found an array with shape: {array.shape}")

    tasks = [
        Task(
            session=session,
            time=time,
            seed=time,
        )
        for time in range(args.t_from, args.t_to, 1)
    ]

    client = Client(n_workers=args.workers, threads_per_worker=1)

    map_result = client.map(execute_read_task, tasks)
    client.gather(map_result)
    print("done, all good")


def main() -> None:
    """Main entry point for the script.

    Parses arguments and delegates to a subcommand.
    """

    global_parser = argparse.ArgumentParser(prog="dask_write")
    global_parser.add_argument(
        "--url",
        type=str,
        help="url for the repository: s3://bucket/optional-prefix/repository-name",
        required=True,
    )
    subparsers = global_parser.add_subparsers(title="subcommands", required=True)

    create_parser = subparsers.add_parser("create", help="create repo and array")
    create_parser.add_argument(
        "--x-chunks", type=int, help="number of chunks in the x dimension", default=4
    )
    create_parser.add_argument(
        "--y-chunks", type=int, help="number of chunks in the y dimension", default=4
    )
    create_parser.add_argument(
        "--t-chunks", type=int, help="number of chunks in the t dimension", default=1000
    )
    create_parser.add_argument(
        "--chunk-x-size",
        type=int,
        help="size of chunks in the x dimension",
        default=112,
    )
    create_parser.add_argument(
        "--chunk-y-size",
        type=int,
        help="size of chunks in the y dimension",
        default=112,
    )
    create_parser.set_defaults(command="create")

    update_parser = subparsers.add_parser("update", help="add chunks to the array")
    update_parser.add_argument(
        "--t-from",
        type=int,
        help="time position where to start adding chunks (included)",
        required=True,
    )
    update_parser.add_argument(
        "--t-to",
        type=int,
        help="time position where to stop adding chunks (not included)",
        required=True,
    )
    update_parser.add_argument(
        "--workers", type=int, help="number of workers to use", required=True
    )
    update_parser.set_defaults(command="update")

    verify_parser = subparsers.add_parser("verify", help="verify array chunks")
    verify_parser.add_argument(
        "--t-from",
        type=int,
        help="time position where to start adding chunks (included)",
        required=True,
    )
    verify_parser.add_argument(
        "--t-to",
        type=int,
        help="time position where to stop adding chunks (not included)",
        required=True,
    )
    verify_parser.add_argument(
        "--workers", type=int, help="number of workers to use", required=True
    )
    verify_parser.set_defaults(command="verify")

    args = global_parser.parse_args()
    url = urlparse(args.url, "s3")
    if (
        url.scheme != "s3"
        or url.netloc == ""
        or url.path == ""
        or url.params != ""
        or url.query != ""
        or url.fragment != ""
    ):
        raise ValueError(f"Invalid url {args.url}")

    args.url = url

    match args.command:
        case "create":
            create(args)
        case "update":
            update(args)
        case "verify":
            verify(args)


if __name__ == "__main__":
    main()



import pytest

import icechunk as ic
import xarray as xr


@pytest.mark.parametrize(
    "storage",
    [
        pytest.param(
            ic.r2_storage(
                prefix="v1/era5_weatherbench2",
                endpoint_url="https://data.icechunk.cloud",
                anonymous=True,
            ),
            id="public-url",
        ),
        pytest.param(
            ic.s3_storage(
                bucket="icechunk-public-data-r2",
                prefix="v1/era5_weatherbench2",
                endpoint_url="https://caa3022c13c9823de0d22b3b6c249494.r2.cloudflarestorage.com",
                region="auto",
                force_path_style=True,
            ),
            id="full-s3",
            marks=pytest.mark.skip(),
        ),
        pytest.param(
            ic.r2_storage(
                bucket="icechunk-public-data-r2",
                prefix="v1/era5_weatherbench2",
                endpoint_url="https://caa3022c13c9823de0d22b3b6c249494.r2.cloudflarestorage.com",
            ),
            id="endpoint-bucket",
            marks=pytest.mark.skip(),
        ),
        pytest.param(
            ic.r2_storage(
                bucket="icechunk-public-data-r2",
                prefix="v1/era5_weatherbench2",
                account_id="caa3022c13c9823de0d22b3b6c249494",
            ),
            id="account-bucket",
            marks=pytest.mark.skip(),
        ),
    ],
)
def test_r2(storage):
    repo = ic.Repository.open(storage=storage)
    session = repo.readonly_session("main")

    # simply test that we can open
    xr.open_dataset(
        session.store,
        group="1x721x1440",
        engine="zarr",
        chunks=None,
        consolidated=False,
    )



"""This test reads a repository generated with an older version of icechunk.

In this way, we check we maintain read compatibility. The repository lives
in the git repository, as a filesystem store, in the directory icechunk-python/tests/data/test-repo

If something changes in the on disk format, we probably won't be able to read the repo, the
test will fail and we can avoid breaking user data.

When new features that impact the stored info are added, or when the on-disk format is
intentionally changed, the repository files must be regenerated. For that, run the current
file as a python script: `python ./tests/test_can_read_old.py`.
"""

from datetime import UTC, datetime
from typing import cast

import pytest
from numpy.testing import assert_array_equal

import icechunk as ic
import zarr

UPDATED_SPLITTING_CONFIG = ic.ManifestSplittingConfig.from_dict(
    {
        ic.ManifestSplitCondition.name_matches("split_*"): {
            ic.ManifestSplitDimCondition.Axis(0): 1,
            ic.ManifestSplitDimCondition.DimensionName("longitude"): 1,
            ic.ManifestSplitDimCondition.Any(): 3,
        }
    }
)


def mk_repo(
    *, create: bool, store_path: str, config: ic.RepositoryConfig | None = None
) -> ic.Repository:
    """Create a store that can access virtual chunks in localhost MinIO"""
    if create and config is None:
        config = ic.RepositoryConfig.default()
        config.inline_chunk_threshold_bytes = 12

        virtual_store_config = ic.s3_store(
            region="us-east-1",
            endpoint_url="http://localhost:9000",
            allow_http=True,
            s3_compatible=True,
            force_path_style=True,
        )
        container = ic.VirtualChunkContainer("s3://testbucket/", virtual_store_config)
        config.set_virtual_chunk_container(container)
    credentials = ic.containers_credentials(
        {
            "s3://testbucket": ic.s3_credentials(
                access_key_id="minio123", secret_access_key="minio123"
            )
        }
    )

    operation = ic.Repository.create if create else ic.Repository.open
    repo = operation(
        storage=ic.local_filesystem_storage(store_path),
        config=config,
        authorize_virtual_chunk_access=credentials,
    )

    return repo


async def write_a_split_repo() -> None:
    """Write the test repository with manifest splitting."""

    store_path = "./tests/data/split-repo"
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 12
    config.manifest = ic.ManifestConfig(
        splitting=ic.ManifestSplittingConfig.from_dict(
            {
                ic.ManifestSplitCondition.name_matches("split_*"): {
                    ic.ManifestSplitDimCondition.Axis(0): 1,
                    ic.ManifestSplitDimCondition.DimensionName("latitude"): 1,
                    ic.ManifestSplitDimCondition.Any(): 3,
                }
            }
        )
    )

    print(f"Writing repository to {store_path}")
    repo = mk_repo(create=True, store_path=store_path, config=config)
    session = repo.writable_session("main")
    store = session.store

    root = zarr.group(store=store)
    group1 = root.create_group(
        "group1", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )

    # these chunks will be materialized
    big_chunks = group1.create_array(
        "split",
        shape=(10, 10),
        chunks=(3, 3),
        dimension_names=(None, "longitude"),
        dtype="float32",
        fill_value=float("nan"),
        attributes={"this": "is a nice array", "icechunk": 1, "size": 42.0},
    )

    # these chunks will be inline
    small_chunks = group1.create_array(
        "small_chunks",
        shape=(5,),
        chunks=(1,),
        dtype="int8",
        fill_value=8,
        attributes={"this": "is a nice array", "icechunk": 1, "size": 42.0},
    )
    session.commit("empty structure")

    session = repo.writable_session("main")
    big_chunks = zarr.open_array(session.store, path="/group1/split", mode="a")
    small_chunks = zarr.open_array(session.store, path="/group1/small_chunks", mode="a")
    big_chunks[:] = 120
    small_chunks[:] = 0
    session.commit("write data")

    session = repo.writable_session("main")
    big_chunks = zarr.open_array(session.store, path="group1/split", mode="a")
    small_chunks = zarr.open_array(session.store, path="group1/small_chunks", mode="a")
    big_chunks[:] = 12
    small_chunks[:] = 1
    session.commit("write data again")

    ### new config
    config = ic.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 12
    config.manifest = ic.ManifestConfig(splitting=UPDATED_SPLITTING_CONFIG)
    repo = mk_repo(create=False, store_path=store_path, config=config)
    repo.save_config()
    session = repo.writable_session("main")
    big_chunks = zarr.open_array(session.store, path="group1/split", mode="a")
    small_chunks = zarr.open_array(session.store, path="group1/small_chunks", mode="a")
    big_chunks[:] = 14
    small_chunks[:] = 3
    session.commit("write data again with more splits")


async def test_icechunk_can_read_old_repo_with_manifest_splitting() -> None:
    repo = mk_repo(create=False, store_path="./tests/data/split-repo")
    ancestry = list(repo.ancestry(branch="main"))[::-1]

    init_snapshot = ancestry[1]
    assert init_snapshot.message == "empty structure"
    assert len(init_snapshot.manifests) == 0

    snapshot = ancestry[2]
    assert snapshot.message == "write data"
    assert len(snapshot.manifests) == 9

    snapshot = ancestry[3]
    assert snapshot.message == "write data again"
    assert len(snapshot.manifests) == 9

    snapshot = ancestry[4]
    assert snapshot.message == "write data again with more splits"
    assert len(snapshot.manifests) == 17

    assert repo.config.manifest
    assert repo.config.manifest.splitting == UPDATED_SPLITTING_CONFIG


async def write_a_test_repo() -> None:
    """Write the test repository.

    This function tries to explore as many icechunk features as possible, to generate
    an richer repository on disk. For example, it does several commits, it has a hierarchy,
    it has virtual, inline and materialized chunks, branches and tags, etc.

    PLEASE: keep addign more actions to this function as we add more features to Icechunk.
    """

    print("Writing repository to ./tests/data/test-repo")
    repo = mk_repo(create=True, store_path="./tests/data/test-repo")
    session = repo.writable_session("main")
    store = session.store

    root = zarr.group(store=store)
    group1 = root.create_group(
        "group1", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )

    # these chunks will be materialized
    big_chunks = group1.create_array(
        "big_chunks",
        shape=(10, 10),
        chunks=(5, 5),
        dtype="float32",
        fill_value=float("nan"),
        attributes={"this": "is a nice array", "icechunk": 1, "size": 42.0},
    )

    # these chunks will be inline
    small_chunks = group1.create_array(
        "small_chunks",
        shape=(5,),
        chunks=(1,),
        dtype="int8",
        fill_value=8,
        attributes={"this": "is a nice array", "icechunk": 1, "size": 42.0},
    )
    session.commit("empty structure")

    session = repo.writable_session("main")
    store = session.store
    root = zarr.group(store=store)
    big_chunks = cast(zarr.Array, root["group1/big_chunks"])
    small_chunks = cast(zarr.Array, root["group1/small_chunks"])

    big_chunks[:] = 42.0
    small_chunks[:] = 84
    snapshot = session.commit("fill data")
    session = repo.writable_session("main")
    store = session.store

    # We are going to write this chunk to storage as a virtual chunk
    # We need to know its size
    buffer_prototype = zarr.core.buffer.default_buffer_prototype()
    chunk = await store.get("group1/big_chunks/c/0/1", prototype=buffer_prototype)
    assert chunk
    virtual_chunk_data_size = len(chunk.to_bytes())

    store.set_virtual_ref(
        "group1/big_chunks/c/0/0",
        "s3://testbucket/can_read_old/chunk-1",
        offset=0,
        length=virtual_chunk_data_size,
        checksum=datetime(9999, 12, 31, tzinfo=UTC),
    )
    snapshot = session.commit("set virtual chunk")
    session = repo.writable_session("main")
    store = session.store

    repo.create_branch("my-branch", snapshot_id=snapshot)
    session = repo.writable_session("my-branch")
    store = session.store

    await store.delete("group1/small_chunks/c/4")
    snap4 = session.commit("delete a chunk")

    repo.create_tag("it works!", snapshot_id=snap4)
    repo.create_tag("deleted", snapshot_id=snap4)
    repo.delete_tag("deleted")

    session = repo.writable_session("my-branch")
    store = session.store
    root = zarr.open_group(store=store)

    group2 = root.create_group(
        "group2", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )
    group3 = group2.create_group(
        "group3", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )
    group4 = group3.create_group(
        "group4", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )
    group5 = group4.create_group(
        "group5", attributes={"this": "is a nice group", "icechunk": 1, "size": 42.0}
    )
    group5.create_array(
        "inner",
        shape=(10, 10),
        chunks=(5, 5),
        dtype="float32",
        fill_value=float("nan"),
        attributes={"this": "is a nice array", "icechunk": 1, "size": 42.0},
    )
    snap5 = session.commit("some more structure")

    repo.create_tag("it also works!", snapshot_id=snap5)

    store.close()


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow")
async def test_icechunk_can_read_old_repo() -> None:
    # we import here so it works when the script is ran by pytest
    from tests.conftest import write_chunks_to_minio

    repo = mk_repo(create=False, store_path="./tests/data/test-repo")
    main_snapshot = repo.lookup_branch("main")

    expected_main_history = [
        "set virtual chunk",
        "fill data",
        "empty structure",
        "Repository initialized",
    ]
    assert [
        p.message for p in repo.ancestry(snapshot_id=main_snapshot)
    ] == expected_main_history

    expected_branch_history = [
        "some more structure",
        "delete a chunk",
    ] + expected_main_history

    assert [
        p.message for p in repo.ancestry(branch="my-branch")
    ] == expected_branch_history

    assert [
        p.message for p in repo.ancestry(tag="it also works!")
    ] == expected_branch_history

    assert [p.message for p in repo.ancestry(tag="it works!")] == expected_branch_history[
        1:
    ]

    with pytest.raises(ic.IcechunkError, match="ref not found"):
        repo.readonly_session(tag="deleted")

    session = repo.writable_session("my-branch")
    store = session.store
    assert sorted([p async for p in store.list_dir("")]) == [
        "group1",
        "group2",
        "zarr.json",
    ]
    assert sorted([p async for p in store.list_dir("group1")]) == [
        "big_chunks",
        "small_chunks",
        "zarr.json",
    ]
    assert sorted([p async for p in store.list_dir("group2")]) == ["group3", "zarr.json"]
    assert sorted([p async for p in store.list_dir("group2/group3")]) == [
        "group4",
        "zarr.json",
    ]
    assert sorted([p async for p in store.list_dir("group2/group3/group4")]) == [
        "group5",
        "zarr.json",
    ]
    assert sorted([p async for p in store.list_dir("group2/group3/group4/group5")]) == [
        "inner",
        "zarr.json",
    ]
    assert sorted(
        [p async for p in store.list_dir("group2/group3/group4/group5/inner")]
    ) == ["c", "zarr.json"]

    root = zarr.group(store=store)
    # inner is not initialized, so it's all fill values
    inner = root["group2/group3/group4/group5/inner"]
    assert_array_equal(inner[:], float("nan"))

    small_chunks = root["group1/small_chunks"]
    # has 5 elements, we deleted the last chunk (of size 1), and the fill value is 8
    assert_array_equal(small_chunks[:], [84, 84, 84, 84, 8])

    # big_chunks array has a virtual chunk, so we need to write it to local MinIO
    # we get the bytes from one of the materialized chunks
    buffer_prototype = zarr.core.buffer.default_buffer_prototype()
    chunk_data = await store.get("group1/big_chunks/c/0/1", prototype=buffer_prototype)
    assert chunk_data
    chunk_data_bytes = chunk_data.to_bytes()

    # big chunks array has a virtual chunk pointing here
    write_chunks_to_minio(
        [
            ("can_read_old/chunk-1", chunk_data_bytes),
        ]
    )

    big_chunks = root["group1/big_chunks"]
    assert_array_equal(big_chunks[:], 42.0)

    parents = list(repo.ancestry(branch="main"))
    diff = repo.diff(to_branch="main", from_snapshot_id=parents[-2].id)
    assert diff.new_groups == set()
    assert diff.new_arrays == set()
    assert set(diff.updated_chunks.keys()) == {
        "/group1/big_chunks",
        "/group1/small_chunks",
    }
    assert sorted(diff.updated_chunks["/group1/big_chunks"]) == sorted(
        [[i, j] for i in range(2) for j in range(2)]
    )
    assert sorted(diff.updated_chunks["/group1/small_chunks"]) == sorted(
        [[i] for i in range(5)]
    )
    assert diff.deleted_groups == set()
    assert diff.deleted_arrays == set()
    assert diff.updated_groups == set()
    assert diff.updated_arrays == set()


if __name__ == "__main__":
    import asyncio

    # we import here so it works when the script is ran by pytest

    # asyncio.run(write_a_test_repo())
    asyncio.run(write_a_split_repo())



import os
import re
from collections.abc import Generator
from pathlib import Path

import pytest

import icechunk
import zarr


@pytest.fixture(scope="function")
def tmp_store(tmpdir: Path) -> Generator[tuple[icechunk.IcechunkStore, str]]:
    repo_path = f"{tmpdir}"
    config = icechunk.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 5
    repo = icechunk.Repository.open_or_create(
        storage=icechunk.local_filesystem_storage(repo_path),
        config=config,
    )

    session = repo.writable_session("main")
    store = session.store

    yield store, repo_path


def test_config_fetch() -> None:
    config = icechunk.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 5
    storage = icechunk.in_memory_storage()
    repo = icechunk.Repository.create(
        storage=storage,
        config=config,
    )

    assert repo.config == config
    assert icechunk.Repository.fetch_config(storage) == config


def test_config_save() -> None:
    config = icechunk.RepositoryConfig.default()
    storage = icechunk.in_memory_storage()
    repo = icechunk.Repository.create(
        storage=storage,
    )

    config.inline_chunk_threshold_bytes = 5
    repo = icechunk.Repository.open(
        storage=storage,
        config=config,
    )

    assert icechunk.Repository.fetch_config(storage) != config
    repo.save_config()
    assert icechunk.Repository.fetch_config(storage) == config


def test_no_inline_chunks(tmp_store: tuple[icechunk.IcechunkStore, str]) -> None:
    store = tmp_store[0]
    store_path = tmp_store[1]
    array = zarr.create_array(
        store=store,
        shape=(10),
        dtype="int64",
        zarr_format=3,
        chunks=(1,),
        fill_value=-1,
    )
    array[:] = 42

    # Check that the chunks directory was created, since each chunk is 4 bytes and the
    # inline_chunk_threshold is 1, we should have 10 chunks in the chunks directory
    assert os.path.isdir(f"{store_path}/chunks")
    assert len(os.listdir(f"{store_path}/chunks")) == 10


def test_inline_chunks(tmp_store: tuple[icechunk.IcechunkStore, str]) -> None:
    store = tmp_store[0]
    store_path = tmp_store[1]

    inline_array = zarr.create_array(
        store=store,
        name="inline",
        shape=(10),
        dtype="int32",
        zarr_format=3,
        chunks=(1,),
        fill_value=-1,
        compressors=None,
    )

    inline_array[:] = 9

    # Check that the chunks directory was not created, since each chunk is 4 bytes and the
    # inline_chunk_threshold is 40, we should have no chunks directory
    assert not os.path.isdir(f"{store_path}/chunks")

    written_array = zarr.create_array(
        store=store,
        name="not_inline",
        shape=(10),
        dtype="int64",
        zarr_format=3,
        chunks=(1,),
        fill_value=-1,
        compressors=None,
    )

    written_array[:] = 3

    # Check that the chunks directory was not created, since each chunk is 8 bytes and the
    # inline_chunk_threshold is 40, we should have 10 chunks in the chunks directory
    assert os.path.isdir(f"{store_path}/chunks")
    assert len(os.listdir(f"/{store_path}/chunks")) == 10


def test_virtual_chunk_containers() -> None:
    config = icechunk.RepositoryConfig.default()

    store_config = icechunk.s3_store(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        s3_compatible=True,
    )
    container = icechunk.VirtualChunkContainer("s3://testbucket/", store_config)
    config.set_virtual_chunk_container(container)
    assert re.match(
        r"RepositoryConfig\(inline_chunk_threshold_bytes=None, get_partial_values_concurrency=None, compression=None, caching=None, storage=None, manifest=.*\)",
        repr(config),
    )
    assert config.virtual_chunk_containers
    assert len(config.virtual_chunk_containers) == 1
    assert config.virtual_chunk_containers["s3://testbucket/"] == container

    config.clear_virtual_chunk_containers()
    assert {} == config.virtual_chunk_containers

    config.set_virtual_chunk_container(container)
    assert len(config.virtual_chunk_containers) == 1
    assert config.virtual_chunk_containers["s3://testbucket/"] == container


def test_can_change_deep_config_values() -> None:
    storage = icechunk.in_memory_storage()
    repo = icechunk.Repository.create(
        storage=storage,
    )
    config = icechunk.RepositoryConfig(
        inline_chunk_threshold_bytes=11,
        compression=icechunk.CompressionConfig(level=0),
    )
    config.inline_chunk_threshold_bytes = 5
    config.get_partial_values_concurrency = 42
    config.compression = icechunk.CompressionConfig(level=8)
    config.compression.level = 2
    config.caching = icechunk.CachingConfig(num_chunk_refs=8)
    config.storage = storage.default_settings()
    assert config.storage.concurrency
    config.storage.concurrency.ideal_concurrent_request_size = 1_000_000
    assert config.storage.retries
    config.storage.retries.max_tries = 42
    config.storage.retries.initial_backoff_ms = 500
    config.storage.retries.max_backoff_ms = 600
    config.storage.storage_class = "STANDARD_IA"
    config.manifest = icechunk.ManifestConfig()
    config.manifest.preload = icechunk.ManifestPreloadConfig(max_total_refs=42)
    config.max_concurrent_requests = 10
    config.manifest.preload.preload_if = icechunk.ManifestPreloadCondition.and_conditions(
        [
            icechunk.ManifestPreloadCondition.true(),
            icechunk.ManifestPreloadCondition.name_matches("foo"),
        ]
    )

    assert re.match(
        r"RepositoryConfig\(inline_chunk_threshold_bytes=5, get_partial_values_concurrency=42, compression=CompressionConfig\(algorithm=None, level=2\), caching=CachingConfig\(num_snapshot_nodes=None, num_chunk_refs=8, num_transaction_changes=None, num_bytes_attributes=None, num_bytes_chunks=None\), storage=StorageSettings\(concurrency=StorageConcurrencySettings\(max_concurrent_requests_for_object=5, ideal_concurrent_request_size=1000000\), retries=StorageRetriesSettings\(max_tries=42, initial_backoff_ms=500, max_backoff_ms=600\), unsafe_use_conditional_create=None, unsafe_use_conditional_update=None, unsafe_use_metadata=None, storage_class=\"STANDARD_IA\", metadata_storage_class=None, chunks_storage_class=None\), manifest=.*\)",
        repr(config),
    )
    repo = icechunk.Repository.open(
        storage=storage,
        config=config,
    )
    repo.save_config()

    stored_config = icechunk.Repository.fetch_config(storage)
    assert stored_config
    assert stored_config.inline_chunk_threshold_bytes == 5
    assert stored_config.compression
    assert stored_config.compression.level == 2
    assert stored_config.max_concurrent_requests == 10
    assert stored_config.caching
    assert stored_config.caching.num_chunk_refs == 8
    assert stored_config.storage
    assert stored_config.storage.concurrency
    assert stored_config.storage.concurrency.ideal_concurrent_request_size == 1_000_000
    assert stored_config.storage.retries
    assert stored_config.storage.retries.max_tries == 42
    assert stored_config.storage.retries.initial_backoff_ms == 500
    assert stored_config.storage.retries.max_backoff_ms == 600
    assert stored_config.manifest
    assert stored_config.manifest.preload
    assert config.manifest.preload.max_total_refs == 42
    assert (
        config.manifest.preload.preload_if
        == icechunk.ManifestPreloadCondition.and_conditions(
            [
                icechunk.ManifestPreloadCondition.true(),
                icechunk.ManifestPreloadCondition.name_matches("foo"),
            ]
        )
    )


def test_manifest_preload_magic_methods() -> None:
    assert (
        icechunk.ManifestPreloadCondition.and_conditions(
            [
                icechunk.ManifestPreloadCondition.true(),
                icechunk.ManifestPreloadCondition.name_matches("foo"),
            ]
        )
        == icechunk.ManifestPreloadCondition.true()
        & icechunk.ManifestPreloadCondition.name_matches("foo")
    )
    assert (
        icechunk.ManifestPreloadCondition.or_conditions(
            [
                icechunk.ManifestPreloadCondition.true(),
                icechunk.ManifestPreloadCondition.name_matches("foo"),
            ]
        )
        == icechunk.ManifestPreloadCondition.true()
        | icechunk.ManifestPreloadCondition.name_matches("foo")
    )


def test_spec_version() -> None:
    assert icechunk.spec_version() >= 1


def test_config_from_store() -> None:
    config = icechunk.RepositoryConfig.default()
    config.inline_chunk_threshold_bytes = 5

    store_config = icechunk.s3_store(region="us-east-1")
    container = icechunk.VirtualChunkContainer("s3://example/", store_config)
    config.set_virtual_chunk_container(container)

    storage = icechunk.in_memory_storage()
    repo = icechunk.Repository.create(
        storage=storage,
        config=config,
    )
    session = repo.writable_session("main")
    store = session.store

    assert store.session.config == config
    assert store.session.config.virtual_chunk_containers.keys() == {"s3://example/"}


def test_s3_storage_options() -> None:
    _storage = icechunk.s3_storage(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        force_path_style=True,
        bucket="testbucket",
        prefix="this-repo-does-not-exist",
        scatter_initial_credentials=True,
        network_stream_timeout_seconds=50,
    )
    # TODO: add accessors and verify values
    # currently this is only testing we can construct



import abc
import datetime
from collections.abc import AsyncGenerator, AsyncIterator, Mapping
from enum import Enum
from typing import Any, TypeAlias

class S3Options:
    """Options for accessing an S3-compatible storage backend"""
    def __init__(
        self,
        region: str | None = None,
        endpoint_url: str | None = None,
        allow_http: bool = False,
        anonymous: bool = False,
        force_path_style: bool = False,
        network_stream_timeout_seconds: int | None = None,
    ) -> None:
        """
        Create a new `S3Options` object

        Parameters
        ----------
        region: str | None
            Optional, the region to use for the storage backend.
        endpoint_url: str | None
            Optional, the endpoint URL to use for the storage backend.
        allow_http: bool
            Whether to allow HTTP requests to the storage backend.
        anonymous: bool
            Whether to use anonymous credentials to the storage backend. When `True`, the s3 requests will not be signed.
        force_path_style: bool
            Whether to force use of path-style addressing for buckets.
        network_stream_timeout_seconds: int | None
            Timeout requests if no bytes can be transmitted during this period of time.
            If set to 0, timeout is disabled. Default is 60 seconds.
        """

    @property
    def region(self) -> str | None:
        """
        Optional region to use for the storage backend.

        Returns
        -------
        str | None
            The region configured for the storage backend.
        """
        ...

    @region.setter
    def region(self, value: str | None) -> None:
        """
        Set the region to use for the storage backend.

        Parameters
        ----------
        value: str | None
            The region to use for the storage backend.
        """
        ...

    @property
    def endpoint_url(self) -> str | None:
        """
        Optional endpoint URL for the storage backend.

        Returns
        -------
        str | None
            The endpoint URL configured for the storage backend.
        """
        ...

    @endpoint_url.setter
    def endpoint_url(self, value: str | None) -> None:
        """
        Set the endpoint URL for the storage backend.

        Parameters
        ----------
        value: str | None
            The endpoint URL to use for the storage backend.
        """
        ...

    @property
    def allow_http(self) -> bool:
        """
        Whether HTTP requests are allowed for the storage backend.

        Returns
        -------
        bool
            ``True`` when HTTP requests to the storage backend are permitted.
        """
        ...

    @allow_http.setter
    def allow_http(self, value: bool) -> None:
        """
        Set whether HTTP requests are allowed for the storage backend.

        Parameters
        ----------
        value: bool
            ``True`` to allow HTTP requests to the storage backend, ``False`` otherwise.
        """
        ...

    @property
    def anonymous(self) -> bool:
        """
        Whether to use anonymous credentials (unsigned requests).

        Returns
        -------
        bool
            ``True`` when anonymous access is configured.
        """
        ...

    @anonymous.setter
    def anonymous(self, value: bool) -> None:
        """
        Set whether to use anonymous credentials.

        Parameters
        ----------
        value: bool
            ``True`` to perform unsigned requests, ``False`` to sign requests.
        """
        ...

    @property
    def force_path_style(self) -> bool:
        """
        Whether to force path-style bucket addressing.

        Returns
        -------
        bool
            ``True`` when path-style addressing is forced.
        """
        ...

    @force_path_style.setter
    def force_path_style(self, value: bool) -> None:
        """
        Set whether to force path-style bucket addressing.

        Parameters
        ----------
        value: bool
            ``True`` to always use path-style addressing, ``False`` to allow virtual-host style.
        """
        ...

    @property
    def network_stream_timeout_seconds(self) -> int | None:
        """
        Timeout in seconds for idle network streams.

        Returns
        -------
        int | None
            The timeout duration; ``0`` disables the timeout and ``None`` uses the default.
        """
        ...

    @network_stream_timeout_seconds.setter
    def network_stream_timeout_seconds(self, value: int | None) -> None:
        """
        Set the timeout for idle network streams.

        Parameters
        ----------
        value: int | None
            Timeout duration in seconds. Use ``0`` to disable or ``None`` for the default.
        """
        ...

class ObjectStoreConfig:
    class InMemory:
        def __init__(self) -> None: ...

    class LocalFileSystem:
        def __init__(self, path: str) -> None: ...

    class S3Compatible:
        def __init__(self, options: S3Options) -> None: ...

    class S3:
        def __init__(self, options: S3Options) -> None: ...

    class Gcs:
        def __init__(self, opts: Mapping[str, str] | None = None) -> None: ...

    class Azure:
        def __init__(self, opts: Mapping[str, str] | None = None) -> None: ...

    class Tigris:
        def __init__(self, opts: S3Options) -> None: ...

    class Http:
        def __init__(self, opts: Mapping[str, str] | None = None) -> None: ...

AnyObjectStoreConfig = (
    ObjectStoreConfig.InMemory
    | ObjectStoreConfig.LocalFileSystem
    | ObjectStoreConfig.S3
    | ObjectStoreConfig.S3Compatible
    | ObjectStoreConfig.Gcs
    | ObjectStoreConfig.Azure
    | ObjectStoreConfig.Tigris
    | ObjectStoreConfig.Http
)

class VirtualChunkContainer:
    """A virtual chunk container is a configuration that allows Icechunk to read virtual references from a storage backend.

    Attributes
    ----------
    url_prefix: str
        The prefix of urls that will use this containers configuration for reading virtual references.
    store: ObjectStoreConfig
        The storage backend to use for the virtual chunk container.
    """

    name: str
    url_prefix: str
    store: ObjectStoreConfig

    def __init__(self, url_prefix: str, store: AnyObjectStoreConfig):
        """
        Create a new `VirtualChunkContainer` object

        Parameters
        ----------
        url_prefix: str
            The prefix of urls that will use this containers configuration for reading virtual references.
        store: ObjectStoreConfig
            The storage backend to use for the virtual chunk container.
        """

class VirtualChunkSpec:
    """The specification for a virtual chunk reference."""
    @property
    def index(self) -> list[int]:
        """The chunk index, in chunk coordinates space"""
        ...
    @property
    def location(self) -> str:
        """The URL to the virtual chunk data, something like 's3://bucket/foo.nc'"""
        ...
    @property
    def offset(self) -> int:
        """The chunk offset within the pointed object, in bytes"""
        ...
    @property
    def length(self) -> int:
        """The length of the chunk in bytes"""
        ...
    @property
    def etag_checksum(self) -> str | None:
        """Optional object store e-tag for the containing object.

        Icechunk will refuse to serve data from this chunk if the etag has changed.
        """
        ...
    @property
    def last_updated_at_checksum(self) -> datetime.datetime | None:
        """Optional timestamp for the containing object.

        Icechunk will refuse to serve data from this chunk if it has been modified in object store after this time.
        """
        ...

    def __init__(
        self,
        index: list[int],
        location: str,
        offset: int,
        length: int,
        etag_checksum: str | None = None,
        last_updated_at_checksum: datetime.datetime | None = None,
    ) -> None: ...

class CompressionAlgorithm(Enum):
    """Enum for selecting the compression algorithm used by Icechunk to write its metadata files

    Attributes
    ----------
    Zstd: int
        The Zstd compression algorithm.
    """

    Zstd = 0

    def __init__(self) -> None: ...
    @staticmethod
    def default() -> CompressionAlgorithm:
        """
        The default compression algorithm used by Icechunk to write its metadata files.

        Returns
        -------
        CompressionAlgorithm
            The default compression algorithm.
        """
        ...

class CompressionConfig:
    """Configuration for how Icechunk compresses its metadata files"""

    def __init__(
        self, algorithm: CompressionAlgorithm | None = None, level: int | None = None
    ) -> None:
        """
        Create a new `CompressionConfig` object

        Parameters
        ----------
        algorithm: CompressionAlgorithm | None
            The compression algorithm to use.
        level: int | None
            The compression level to use.
        """
        ...
    @property
    def algorithm(self) -> CompressionAlgorithm | None:
        """
        The compression algorithm used by Icechunk to write its metadata files.

        Returns
        -------
        CompressionAlgorithm | None
            The compression algorithm used by Icechunk to write its metadata files.
        """
        ...
    @algorithm.setter
    def algorithm(self, value: CompressionAlgorithm | None) -> None:
        """
        Set the compression algorithm used by Icechunk to write its metadata files.

        Parameters
        ----------
        value: CompressionAlgorithm | None
            The compression algorithm to use.
        """
        ...
    @property
    def level(self) -> int | None:
        """
        The compression level used by Icechunk to write its metadata files.

        Returns
        -------
        int | None
            The compression level used by Icechunk to write its metadata files.
        """
        ...
    @level.setter
    def level(self, value: int | None) -> None:
        """
        Set the compression level used by Icechunk to write its metadata files.

        Parameters
        ----------
        value: int | None
            The compression level to use.
        """
        ...
    @staticmethod
    def default() -> CompressionConfig:
        """
        The default compression configuration used by Icechunk to write its metadata files.

        Returns
        -------
        CompressionConfig
        """

class CachingConfig:
    """Configuration for how Icechunk caches its metadata files"""

    def __init__(
        self,
        num_snapshot_nodes: int | None = None,
        num_chunk_refs: int | None = None,
        num_transaction_changes: int | None = None,
        num_bytes_attributes: int | None = None,
        num_bytes_chunks: int | None = None,
    ) -> None:
        """
        Create a new `CachingConfig` object

        Parameters
        ----------
        num_snapshot_nodes: int | None
            The number of snapshot nodes to cache.
        num_chunk_refs: int | None
            The number of chunk references to cache.
        num_transaction_changes: int | None
            The number of transaction changes to cache.
        num_bytes_attributes: int | None
            The number of bytes of attributes to cache.
        num_bytes_chunks: int | None
            The number of bytes of chunks to cache.
        """
    @property
    def num_snapshot_nodes(self) -> int | None:
        """
        The number of snapshot nodes to cache.

        Returns
        -------
        int | None
            The number of snapshot nodes to cache.
        """
        ...
    @num_snapshot_nodes.setter
    def num_snapshot_nodes(self, value: int | None) -> None:
        """
        Set the number of snapshot nodes to cache.

        Parameters
        ----------
        value: int | None
            The number of snapshot nodes to cache.
        """
        ...
    @property
    def num_chunk_refs(self) -> int | None:
        """
        The number of chunk references to cache.

        Returns
        -------
        int | None
            The number of chunk references to cache.
        """
        ...
    @num_chunk_refs.setter
    def num_chunk_refs(self, value: int | None) -> None:
        """
        Set the number of chunk references to cache.

        Parameters
        ----------
        value: int | None
            The number of chunk references to cache.
        """
        ...
    @property
    def num_transaction_changes(self) -> int | None:
        """
        The number of transaction changes to cache.

        Returns
        -------
        int | None
            The number of transaction changes to cache.
        """
        ...
    @num_transaction_changes.setter
    def num_transaction_changes(self, value: int | None) -> None:
        """
        Set the number of transaction changes to cache.

        Parameters
        ----------
        value: int | None
            The number of transaction changes to cache.
        """
        ...
    @property
    def num_bytes_attributes(self) -> int | None:
        """
        The number of bytes of attributes to cache.

        Returns
        -------
        int | None
            The number of bytes of attributes to cache.
        """
        ...
    @num_bytes_attributes.setter
    def num_bytes_attributes(self, value: int | None) -> None:
        """
        Set the number of bytes of attributes to cache.

        Parameters
        ----------
        value: int | None
            The number of bytes of attributes to cache.
        """
        ...
    @property
    def num_bytes_chunks(self) -> int | None:
        """
        The number of bytes of chunks to cache.

        Returns
        -------
        int | None
            The number of bytes of chunks to cache.
        """
        ...
    @num_bytes_chunks.setter
    def num_bytes_chunks(self, value: int | None) -> None:
        """
        Set the number of bytes of chunks to cache.

        Parameters
        ----------
        value: int | None
            The number of bytes of chunks to cache.
        """
        ...

class ManifestPreloadCondition:
    """Configuration for conditions under which manifests will preload on session creation"""

    @staticmethod
    def or_conditions(
        conditions: list[ManifestPreloadCondition],
    ) -> ManifestPreloadCondition:
        """Create a preload condition that matches if any of `conditions` matches"""
        ...
    @staticmethod
    def and_conditions(
        conditions: list[ManifestPreloadCondition],
    ) -> ManifestPreloadCondition:
        """Create a preload condition that matches only if all passed `conditions` match"""
        ...
    @staticmethod
    def path_matches(regex: str) -> ManifestPreloadCondition:
        """Create a preload condition that matches if the full path to the array matches the passed regex.

        Array paths are absolute, as in `/path/to/my/array`
        """
        ...
    @staticmethod
    def name_matches(regex: str) -> ManifestPreloadCondition:
        """Create a preload condition that matches if the array's name matches the passed regex.

        Example, for an array  `/model/outputs/temperature`, the following will match:
        ```
        name_matches(".*temp.*")
        ```
        """
        ...
    @staticmethod
    def num_refs(from_refs: int | None, to_refs: int | None) -> ManifestPreloadCondition:
        """Create a preload condition that matches only if the number of chunk references in the manifest is within the given range.

        from_refs is inclusive, to_refs is exclusive.
        """
        ...
    @staticmethod
    def true() -> ManifestPreloadCondition:
        """Create a preload condition that always matches any manifest"""
        ...
    @staticmethod
    def false() -> ManifestPreloadCondition:
        """Create a preload condition that never matches any manifests"""
        ...
    def __and__(self, other: ManifestPreloadCondition) -> ManifestPreloadCondition:
        """Create a preload condition that matches if both this condition and `other` match."""
        ...
    def __or__(self, other: ManifestPreloadCondition) -> ManifestPreloadCondition:
        """Create a preload condition that matches if either this condition or `other` match."""
        ...

class ManifestPreloadConfig:
    """Configuration for how Icechunk manifest preload on session creation"""

    def __init__(
        self,
        max_total_refs: int | None = None,
        preload_if: ManifestPreloadCondition | None = None,
    ) -> None:
        """
        Create a new `ManifestPreloadConfig` object

        Parameters
        ----------
        max_total_refs: int | None
            The maximum number of references to preload.
        preload_if: ManifestPreloadCondition | None
            The condition under which manifests will be preloaded.
        """
        ...
    @property
    def max_total_refs(self) -> int | None:
        """
        The maximum number of references to preload.

        Returns
        -------
        int | None
            The maximum number of references to preload.
        """
        ...
    @max_total_refs.setter
    def max_total_refs(self, value: int | None) -> None:
        """
        Set the maximum number of references to preload.

        Parameters
        ----------
        value: int | None
            The maximum number of references to preload.
        """
        ...
    @property
    def preload_if(self) -> ManifestPreloadCondition | None:
        """
        The condition under which manifests will be preloaded.

        Returns
        -------
        ManifestPreloadCondition | None
            The condition under which manifests will be preloaded.
        """
        ...
    @preload_if.setter
    def preload_if(self, value: ManifestPreloadCondition | None) -> None:
        """
        Set the condition under which manifests will be preloaded.

        Parameters
        ----------
        value: ManifestPreloadCondition | None
            The condition under which manifests will be preloaded.
        """
        ...

class ManifestSplitCondition:
    """Configuration for conditions under which manifests will be split into splits"""

    @staticmethod
    def or_conditions(
        conditions: list[ManifestSplitCondition],
    ) -> ManifestSplitCondition:
        """Create a splitting condition that matches if any of `conditions` matches"""
        ...
    @staticmethod
    def and_conditions(
        conditions: list[ManifestSplitCondition],
    ) -> ManifestSplitCondition:
        """Create a splitting condition that matches only if all passed `conditions` match"""
        ...
    @staticmethod
    def path_matches(regex: str) -> ManifestSplitCondition:
        """Create a splitting condition that matches if the full path to the array matches the passed regex.

        Array paths are absolute, as in `/path/to/my/array`
        """
        ...
    @staticmethod
    def name_matches(regex: str) -> ManifestSplitCondition:
        """Create a splitting condition that matches if the array's name matches the passed regex.

        Example, for an array  `/model/outputs/temperature`, the following will match:
        ```
        name_matches(".*temp.*")
        ```
        """
        ...

    @staticmethod
    def AnyArray() -> ManifestSplitCondition:
        """Create a splitting condition that matches any array."""
        ...

    def __or__(self, other: ManifestSplitCondition) -> ManifestSplitCondition:
        """Create a splitting condition that matches if either this condition or `other` matches"""
        ...

    def __and__(self, other: ManifestSplitCondition) -> ManifestSplitCondition:
        """Create a splitting condition that matches if both this condition and `other` match"""
        ...

class ManifestSplitDimCondition:
    """Conditions for specifying dimensions along which to shard manifests."""
    class Axis:
        """Split along specified integer axis."""
        def __init__(self, axis: int) -> None: ...

    class DimensionName:
        """Split along specified named dimension."""
        def __init__(self, regex: str) -> None: ...

    class Any:
        """Split along any other unspecified dimension."""
        def __init__(self) -> None: ...

DimSplitSize: TypeAlias = int
SplitSizes: TypeAlias = tuple[
    tuple[
        ManifestSplitCondition,
        tuple[
            tuple[
                ManifestSplitDimCondition.Axis
                | ManifestSplitDimCondition.DimensionName
                | ManifestSplitDimCondition.Any,
                DimSplitSize,
            ],
            ...,
        ],
    ],
    ...,
]

class ManifestSplittingConfig:
    """Configuration for manifest splitting."""

    @staticmethod
    def from_dict(
        split_sizes: dict[
            ManifestSplitCondition,
            dict[
                ManifestSplitDimCondition.Axis
                | ManifestSplitDimCondition.DimensionName
                | ManifestSplitDimCondition.Any,
                int,
            ],
        ],
    ) -> ManifestSplittingConfig: ...
    def __init__(self, split_sizes: SplitSizes) -> None:
        """Configuration for how Icechunk manifests will be split.

        Parameters
        ----------
        split_sizes: tuple[tuple[ManifestSplitCondition, tuple[tuple[ManifestSplitDimCondition, int], ...]], ...]
            The configuration for how Icechunk manifests will be preloaded.

        Examples
        --------

        Split manifests for the `temperature` array, with 3 chunks per shard along the `longitude` dimension.
        >>> ManifestSplittingConfig.from_dict(
        ...     {
        ...         ManifestSplitCondition.name_matches("temperature"): {
        ...             ManifestSplitDimCondition.DimensionName("longitude"): 3
        ...         }
        ...     }
        ... )
        """
        pass

    @property
    def split_sizes(self) -> SplitSizes:
        """
        Configuration for how Icechunk manifests will be split.

        Returns
        -------
        tuple[tuple[ManifestSplitCondition, tuple[tuple[ManifestSplitDimCondition, int], ...]], ...]
            The configuration for how Icechunk manifests will be preloaded.
        """
        ...

    @split_sizes.setter
    def split_sizes(self, value: SplitSizes) -> None:
        """
        Set the sizes for how Icechunk manifests will be split.

        Parameters
        ----------
        value: tuple[tuple[ManifestSplitCondition, tuple[tuple[ManifestSplitDimCondition, int], ...]], ...]
            The configuration for how Icechunk manifests will be preloaded.
        """
        ...

class ManifestConfig:
    """Configuration for how Icechunk manifests"""

    def __init__(
        self,
        preload: ManifestPreloadConfig | None = None,
        splitting: ManifestSplittingConfig | None = None,
    ) -> None:
        """
        Create a new `ManifestConfig` object

        Parameters
        ----------
        preload: ManifestPreloadConfig | None
            The configuration for how Icechunk manifests will be preloaded.
        splitting: ManifestSplittingConfig | None
            The configuration for how Icechunk manifests will be split.
        """
        ...
    @property
    def preload(self) -> ManifestPreloadConfig | None:
        """
        The configuration for how Icechunk manifests will be preloaded.

        Returns
        -------
        ManifestPreloadConfig | None
            The configuration for how Icechunk manifests will be preloaded.
        """
        ...
    @preload.setter
    def preload(self, value: ManifestPreloadConfig | None) -> None:
        """
        Set the configuration for how Icechunk manifests will be preloaded.

        Parameters
        ----------
        value: ManifestPreloadConfig | None
            The configuration for how Icechunk manifests will be preloaded.
        """
        ...

    @property
    def splitting(self) -> ManifestSplittingConfig | None:
        """
        The configuration for how Icechunk manifests will be split.

        Returns
        -------
        ManifestSplittingConfig | None
            The configuration for how Icechunk manifests will be split.
        """
        ...

    @splitting.setter
    def splitting(self, value: ManifestSplittingConfig | None) -> None:
        """
        Set the configuration for how Icechunk manifests will be split.

        Parameters
        ----------
        value: ManifestSplittingConfig | None
            The configuration for how Icechunk manifests will be split.
        """
        ...

class StorageRetriesSettings:
    """Configuration for how Icechunk retries requests.

    Icechunk retries failed requests with an exponential backoff algorithm."""

    def __init__(
        self,
        max_tries: int | None = None,
        initial_backoff_ms: int | None = None,
        max_backoff_ms: int | None = None,
    ) -> None:
        """
        Create a new `StorageRetriesSettings` object

        Parameters
        ----------
        max_tries: int | None
            The maximum number of tries, including the initial one. Set to 1 to disable retries
        initial_backoff_ms: int | None
            The initial backoff duration in milliseconds
        max_backoff_ms: int | None
            The limit to backoff duration in milliseconds
        """
        ...
    @property
    def max_tries(self) -> int | None:
        """
        The maximum number of tries, including the initial one.

        Returns
        -------
        int | None
            The maximum number of tries.
        """
        ...
    @max_tries.setter
    def max_tries(self, value: int | None) -> None:
        """
        Set the maximum number of tries. Set to 1 to disable retries.

        Parameters
        ----------
        value: int | None
            The maximum number of tries
        """
        ...
    @property
    def initial_backoff_ms(self) -> int | None:
        """
        The initial backoff duration in milliseconds.

        Returns
        -------
        int | None
            The initial backoff duration in milliseconds.
        """
        ...
    @initial_backoff_ms.setter
    def initial_backoff_ms(self, value: int | None) -> None:
        """
        Set the initial backoff duration in milliseconds.

        Parameters
        ----------
        value: int | None
            The initial backoff duration in milliseconds.
        """
        ...
    @property
    def max_backoff_ms(self) -> int | None:
        """
        The maximum backoff duration in milliseconds.

        Returns
        -------
        int | None
            The maximum backoff duration in milliseconds.
        """
        ...
    @max_backoff_ms.setter
    def max_backoff_ms(self, value: int | None) -> None:
        """
        Set the maximum backoff duration in milliseconds.

        Parameters
        ----------
        value: int | None
            The maximum backoff duration in milliseconds.
        """
        ...

class StorageConcurrencySettings:
    """Configuration for how Icechunk uses its Storage instance"""

    def __init__(
        self,
        max_concurrent_requests_for_object: int | None = None,
        ideal_concurrent_request_size: int | None = None,
    ) -> None:
        """
        Create a new `StorageConcurrencySettings` object

        Parameters
        ----------
        max_concurrent_requests_for_object: int | None
            The maximum number of concurrent requests for an object.
        ideal_concurrent_request_size: int | None
            The ideal concurrent request size.
        """
        ...
    @property
    def max_concurrent_requests_for_object(self) -> int | None:
        """
        The maximum number of concurrent requests for an object.

        Returns
        -------
        int | None
            The maximum number of concurrent requests for an object.
        """
        ...
    @max_concurrent_requests_for_object.setter
    def max_concurrent_requests_for_object(self, value: int | None) -> None:
        """
        Set the maximum number of concurrent requests for an object.

        Parameters
        ----------
        value: int | None
            The maximum number of concurrent requests for an object.
        """
        ...
    @property
    def ideal_concurrent_request_size(self) -> int | None:
        """
        The ideal concurrent request size.

        Returns
        -------
        int | None
            The ideal concurrent request size.
        """
        ...
    @ideal_concurrent_request_size.setter
    def ideal_concurrent_request_size(self, value: int | None) -> None:
        """
        Set the ideal concurrent request size.

        Parameters
        ----------
        value: int | None
            The ideal concurrent request size.
        """
        ...

class StorageSettings:
    """Configuration for how Icechunk uses its Storage instance"""

    def __init__(
        self,
        concurrency: StorageConcurrencySettings | None = None,
        retries: StorageRetriesSettings | None = None,
        unsafe_use_conditional_create: bool | None = None,
        unsafe_use_conditional_update: bool | None = None,
        unsafe_use_metadata: bool | None = None,
        storage_class: str | None = None,
        metadata_storage_class: str | None = None,
        chunks_storage_class: str | None = None,
        minimum_size_for_multipart_upload: int | None = None,
    ) -> None:
        """
        Create a new `StorageSettings` object

        Parameters
        ----------
        concurrency: StorageConcurrencySettings | None
            The configuration for how Icechunk uses its Storage instance.

        retries: StorageRetriesSettings | None
            The configuration for how Icechunk retries failed requests.

        unsafe_use_conditional_update: bool | None
            If set to False, Icechunk loses some of its consistency guarantees.
            This is only useful in object stores that don't support the feature.
            Use it at your own risk.

        unsafe_use_conditional_create: bool | None
            If set to False, Icechunk loses some of its consistency guarantees.
            This is only useful in object stores that don't support the feature.
            Use at your own risk.

        unsafe_use_metadata: bool | None
            Don't write metadata fields in Icechunk files.
            This is only useful in object stores that don't support the feature.
            Use at your own risk.

        storage_class: str | None
            Store all objects using this object store storage class
            If None the object store default will be used.
            Currently not supported in GCS.
            Example: STANDARD_IA

        metadata_storage_class: str | None
            Store metadata objects using this object store storage class.
            Currently not supported in GCS.
            Defaults to storage_class.

        chunks_storage_class: str | None
            Store chunk objects using this object store storage class.
            Currently not supported in GCS.
            Defaults to storage_class.

        minimum_size_for_multipart_upload: int | None
            Use object store's multipart upload for objects larger than this size in bytes.
            Default: 100 MB if None is passed.
        """
        ...
    @property
    def concurrency(self) -> StorageConcurrencySettings | None:
        """
        The configuration for how much concurrency Icechunk store uses

        Returns
        -------
        StorageConcurrencySettings | None
            The configuration for how Icechunk uses its Storage instance.
        """

    @concurrency.setter
    def concurrency(self, value: StorageConcurrencySettings | None) -> None: ...
    @property
    def retries(self) -> StorageRetriesSettings | None:
        """
        The configuration for how Icechunk retries failed requests.

        Returns
        -------
        StorageRetriesSettings | None
            The configuration for how Icechunk retries failed requests.
        """

    @retries.setter
    def retries(self, value: StorageRetriesSettings | None) -> None: ...
    @property
    def unsafe_use_conditional_update(self) -> bool | None:
        """True if Icechunk will use conditional PUT operations for updates in the object store"""
        ...

    @unsafe_use_conditional_update.setter
    def unsafe_use_conditional_update(self, value: bool) -> None: ...
    @property
    def unsafe_use_conditional_create(self) -> bool | None:
        """True if Icechunk will use conditional PUT operations for creation in the object store"""
        ...

    @unsafe_use_conditional_create.setter
    def unsafe_use_conditional_create(self, value: bool) -> None: ...
    @property
    def unsafe_use_metadata(self) -> bool | None:
        """True if Icechunk will write object metadata in the object store"""
        ...

    @unsafe_use_metadata.setter
    def unsafe_use_metadata(self, value: bool) -> None: ...
    @property
    def storage_class(self) -> str | None:
        """All objects in object store will use this storage class or the default if None"""
        ...

    @storage_class.setter
    def storage_class(self, value: str) -> None: ...
    @property
    def metadata_storage_class(self) -> str | None:
        """Metadata objects in object store will use this storage class or self.storage_class if None"""
        ...

    @metadata_storage_class.setter
    def metadata_storage_class(self, value: str) -> None: ...
    @property
    def chunks_storage_class(self) -> str | None:
        """Chunk objects in object store will use this storage class or self.storage_class if None"""
        ...

    @chunks_storage_class.setter
    def chunks_storage_class(self, value: str) -> None: ...
    @property
    def minimum_size_for_multipart_upload(self) -> int | None:
        """Use object store's multipart upload for objects larger than this size in bytes"""
        ...

    @minimum_size_for_multipart_upload.setter
    def minimum_size_for_multipart_upload(self, value: int) -> None: ...

class RepositoryConfig:
    """Configuration for an Icechunk repository"""

    def __init__(
        self,
        inline_chunk_threshold_bytes: int | None = None,
        get_partial_values_concurrency: int | None = None,
        compression: CompressionConfig | None = None,
        max_concurrent_requests: int | None = None,
        caching: CachingConfig | None = None,
        storage: StorageSettings | None = None,
        virtual_chunk_containers: dict[str, VirtualChunkContainer] | None = None,
        manifest: ManifestConfig | None = None,
    ) -> None:
        """
        Create a new `RepositoryConfig` object

        Parameters
        ----------
        inline_chunk_threshold_bytes: int | None
            The maximum size of a chunk that will be stored inline in the repository.
        get_partial_values_concurrency: int | None
            The number of concurrent requests to make when getting partial values from storage.
        compression: CompressionConfig | None
            The compression configuration for the repository.
        max_concurrent_requests: int | None
            The maximum number of concurrent HTTP requests Icechunk will do for this repo.
            Default is 256.
        caching: CachingConfig | None
            The caching configuration for the repository.
        storage: StorageSettings | None
            The storage configuration for the repository.
        virtual_chunk_containers: dict[str, VirtualChunkContainer] | None
            The virtual chunk containers for the repository.
        manifest: ManifestConfig | None
            The manifest configuration for the repository.
        """
        ...
    @staticmethod
    def default() -> RepositoryConfig:
        """Create a default repository config instance"""
        ...
    @property
    def inline_chunk_threshold_bytes(self) -> int | None:
        """
        The maximum size of a chunk that will be stored inline in the repository. Chunks larger than this size will be written to storage.
        """
        ...
    @inline_chunk_threshold_bytes.setter
    def inline_chunk_threshold_bytes(self, value: int | None) -> None:
        """
        Set the maximum size of a chunk that will be stored inline in the repository. Chunks larger than this size will be written to storage.
        """
        ...
    @property
    def get_partial_values_concurrency(self) -> int | None:
        """
        The number of concurrent requests to make when getting partial values from storage.

        Returns
        -------
        int | None
            The number of concurrent requests to make when getting partial values from storage.
        """
        ...
    @get_partial_values_concurrency.setter
    def get_partial_values_concurrency(self, value: int | None) -> None:
        """
        Set the number of concurrent requests to make when getting partial values from storage.

        Parameters
        ----------
        value: int | None
            The number of concurrent requests to make when getting partial values from storage.
        """
        ...
    @property
    def compression(self) -> CompressionConfig | None:
        """
        The compression configuration for the repository.

        Returns
        -------
        CompressionConfig | None
            The compression configuration for the repository.
        """
        ...
    @compression.setter
    def compression(self, value: CompressionConfig | None) -> None:
        """
        Set the compression configuration for the repository.

        Parameters
        ----------
        value: CompressionConfig | None
            The compression configuration for the repository.
        """
        ...
    @property
    def max_concurrent_requests(self) -> int | None:
        """
        The maximum number of concurrent HTTP requests Icechunk will do for this repo.

        Returns
        -------
        int | None
            The maximum number of concurrent HTTP requests Icechunk will do for this repo.
        """
        ...
    @max_concurrent_requests.setter
    def max_concurrent_requests(self, value: int | None) -> None:
        """
        Set the maximum number of concurrent HTTP requests Icechunk should do for this repo.

        Parameters
        ----------
        value: int | None
            The maximum allowed.
        """
        ...
    @property
    def caching(self) -> CachingConfig | None:
        """
        The caching configuration for the repository.

        Returns
        -------
        CachingConfig | None
            The caching configuration for the repository.
        """
        ...
    @caching.setter
    def caching(self, value: CachingConfig | None) -> None:
        """
        Set the caching configuration for the repository.

        Parameters
        ----------
        value: CachingConfig | None
            The caching configuration for the repository.
        """
        ...
    @property
    def storage(self) -> StorageSettings | None:
        """
        The storage configuration for the repository.

        Returns
        -------
        StorageSettings | None
            The storage configuration for the repository.
        """
        ...
    @storage.setter
    def storage(self, value: StorageSettings | None) -> None:
        """
        Set the storage configuration for the repository.

        Parameters
        ----------
        value: StorageSettings | None
            The storage configuration for the repository.
        """
        ...
    @property
    def manifest(self) -> ManifestConfig | None:
        """
        The manifest configuration for the repository.

        Returns
        -------
        ManifestConfig | None
            The manifest configuration for the repository.
        """
        ...
    @manifest.setter
    def manifest(self, value: ManifestConfig | None) -> None:
        """
        Set the manifest configuration for the repository.

        Parameters
        ----------
        value: ManifestConfig | None
            The manifest configuration for the repository.
        """
        ...
    @property
    def virtual_chunk_containers(self) -> dict[str, VirtualChunkContainer] | None:
        """
        The virtual chunk containers for the repository.

        Returns
        -------
        dict[str, VirtualChunkContainer] | None
            The virtual chunk containers for the repository.
        """
        ...
    def get_virtual_chunk_container(self, name: str) -> VirtualChunkContainer | None:
        """
        Get the virtual chunk container for the repository associated with the given name.

        Parameters
        ----------
        name: str
            The name of the virtual chunk container to get.

        Returns
        -------
        VirtualChunkContainer | None
            The virtual chunk container for the repository associated with the given name.
        """
        ...
    def set_virtual_chunk_container(self, cont: VirtualChunkContainer) -> None:
        """
        Set the virtual chunk container for the repository.

        Parameters
        ----------
        cont: VirtualChunkContainer
            The virtual chunk container to set.
        """
        ...
    def clear_virtual_chunk_containers(self) -> None:
        """
        Clear all virtual chunk containers from the repository.
        """
        ...

class Diff:
    """The result of comparing two snapshots"""
    @property
    def new_groups(self) -> set[str]:
        """
        The groups that were added to the target ref.
        """
        ...
    @property
    def new_arrays(self) -> set[str]:
        """
        The arrays that were added to the target ref.
        """
        ...
    @property
    def deleted_groups(self) -> set[str]:
        """
        The groups that were deleted in the target ref.
        """
        ...
    @property
    def deleted_arrays(self) -> set[str]:
        """
        The arrays that were deleted in the target ref.
        """
        ...
    @property
    def updated_groups(self) -> set[str]:
        """
        The groups that were updated via zarr metadata in the target ref.
        """
        ...
    @property
    def updated_arrays(self) -> set[str]:
        """
        The arrays that were updated via zarr metadata in the target ref.
        """
        ...
    @property
    def updated_chunks(self) -> dict[str, list[list[int]]]:
        """
        The chunks indices that had data updated in the target ref, keyed by the path to the array.
        """
        ...

class GCSummary:
    """Summarizes the results of a garbage collection operation on an icechunk repo"""
    @property
    def bytes_deleted(self) -> int:
        """
        How many bytes were deleted.
        """
        ...
    @property
    def chunks_deleted(self) -> int:
        """
        How many chunks were deleted.
        """
        ...
    @property
    def manifests_deleted(self) -> int:
        """
        How many manifests were deleted.
        """
        ...
    @property
    def snapshots_deleted(self) -> int:
        """
        How many snapshots were deleted.
        """
        ...
    @property
    def attributes_deleted(self) -> int:
        """
        How many attributes were deleted.
        """
        ...
    @property
    def transaction_logs_deleted(self) -> int:
        """
        How many transaction logs were deleted.
        """
        ...

class PyRepository:
    @classmethod
    def create(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @classmethod
    async def create_async(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @classmethod
    def open(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @classmethod
    async def open_async(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @classmethod
    def open_or_create(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @classmethod
    async def open_or_create_async(
        cls,
        storage: Storage,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    @staticmethod
    def exists(storage: Storage) -> bool: ...
    @staticmethod
    async def exists_async(storage: Storage) -> bool: ...
    @classmethod
    def from_bytes(cls, data: bytes) -> PyRepository: ...
    def as_bytes(self) -> bytes: ...
    @staticmethod
    def fetch_config(storage: Storage) -> RepositoryConfig | None: ...
    @staticmethod
    async def fetch_config_async(storage: Storage) -> RepositoryConfig | None: ...
    def save_config(self) -> None: ...
    async def save_config_async(self) -> None: ...
    def config(self) -> RepositoryConfig: ...
    def storage_settings(self) -> StorageSettings: ...
    def storage(self) -> Storage: ...
    @property
    def authorized_virtual_container_prefixes(self) -> set[str]: ...
    def reopen(
        self,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    async def reopen_async(
        self,
        *,
        config: RepositoryConfig | None = None,
        authorize_virtual_chunk_access: dict[str, AnyCredential | None] | None = None,
    ) -> PyRepository: ...
    def set_default_commit_metadata(self, metadata: dict[str, Any]) -> None: ...
    def default_commit_metadata(self) -> dict[str, Any]: ...
    def async_ancestry(
        self,
        *,
        branch: str | None = None,
        tag: str | None = None,
        snapshot_id: str | None = None,
    ) -> AsyncIterator[SnapshotInfo]: ...
    def create_branch(self, branch: str, snapshot_id: str) -> None: ...
    async def create_branch_async(self, branch: str, snapshot_id: str) -> None: ...
    def list_branches(self) -> set[str]: ...
    async def list_branches_async(self) -> set[str]: ...
    def lookup_branch(self, branch: str) -> str: ...
    async def lookup_branch_async(self, branch: str) -> str: ...
    def lookup_snapshot(self, snapshot_id: str) -> SnapshotInfo: ...
    async def lookup_snapshot_async(self, snapshot_id: str) -> SnapshotInfo: ...
    def reset_branch(
        self, branch: str, to_snapshot_id: str, from_snapshot_id: str | None
    ) -> None: ...
    async def reset_branch_async(
        self, branch: str, to_snapshot_id: str, from_snapshot_id: str | None
    ) -> None: ...
    def delete_branch(self, branch: str) -> None: ...
    async def delete_branch_async(self, branch: str) -> None: ...
    def delete_tag(self, tag: str) -> None: ...
    async def delete_tag_async(self, tag: str) -> None: ...
    def create_tag(self, tag: str, snapshot_id: str) -> None: ...
    async def create_tag_async(self, tag: str, snapshot_id: str) -> None: ...
    def list_tags(self) -> set[str]: ...
    async def list_tags_async(self) -> set[str]: ...
    def lookup_tag(self, tag: str) -> str: ...
    async def lookup_tag_async(self, tag: str) -> str: ...
    def diff(
        self,
        from_branch: str | None = None,
        from_tag: str | None = None,
        from_snapshot_id: str | None = None,
        to_branch: str | None = None,
        to_tag: str | None = None,
        to_snapshot_id: str | None = None,
    ) -> Diff: ...
    async def diff_async(
        self,
        from_branch: str | None = None,
        from_tag: str | None = None,
        from_snapshot_id: str | None = None,
        to_branch: str | None = None,
        to_tag: str | None = None,
        to_snapshot_id: str | None = None,
    ) -> Diff: ...
    def readonly_session(
        self,
        branch: str | None = None,
        *,
        tag: str | None = None,
        snapshot_id: str | None = None,
        as_of: datetime.datetime | None = None,
    ) -> PySession: ...
    async def readonly_session_async(
        self,
        branch: str | None = None,
        *,
        tag: str | None = None,
        snapshot_id: str | None = None,
        as_of: datetime.datetime | None = None,
    ) -> PySession: ...
    def writable_session(self, branch: str) -> PySession: ...
    async def writable_session_async(self, branch: str) -> PySession: ...
    def expire_snapshots(
        self,
        older_than: datetime.datetime,
        *,
        delete_expired_branches: bool = False,
        delete_expired_tags: bool = False,
    ) -> set[str]: ...
    async def expire_snapshots_async(
        self,
        older_than: datetime.datetime,
        *,
        delete_expired_branches: bool = False,
        delete_expired_tags: bool = False,
    ) -> set[str]: ...
    def rewrite_manifests(
        self, message: str, *, branch: str, metadata: dict[str, Any] | None = None
    ) -> str: ...
    async def rewrite_manifests_async(
        self, message: str, *, branch: str, metadata: dict[str, Any] | None = None
    ) -> str: ...
    def garbage_collect(
        self,
        delete_object_older_than: datetime.datetime,
        *,
        dry_run: bool = False,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> GCSummary: ...
    async def garbage_collect_async(
        self,
        delete_object_older_than: datetime.datetime,
        *,
        dry_run: bool = False,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> GCSummary: ...
    def total_chunks_storage(
        self,
        *,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> int: ...
    async def total_chunks_storage_async(
        self,
        *,
        max_snapshots_in_memory: int = 50,
        max_compressed_manifest_mem_bytes: int = 512 * 1024 * 1024,
        max_concurrent_manifest_fetches: int = 500,
    ) -> int: ...
    def inspect_snapshot(self, snapshot_id: str, *, pretty: bool = True) -> str: ...
    async def inspect_snapshot_async(
        self, snapshot_id: str, *, pretty: bool = True
    ) -> str: ...

class PySession:
    @classmethod
    def from_bytes(cls, data: bytes) -> PySession: ...
    def __eq__(self, value: object) -> bool: ...
    def as_bytes(self) -> bytes: ...
    @property
    def read_only(self) -> bool: ...
    @property
    def snapshot_id(self) -> str: ...
    @property
    def branch(self) -> str | None: ...
    @property
    def has_uncommitted_changes(self) -> bool: ...
    def status(self) -> Diff: ...
    def discard_changes(self) -> None: ...
    def all_virtual_chunk_locations(self) -> list[str]: ...
    async def all_virtual_chunk_locations_async(self) -> list[str]: ...
    def chunk_coordinates(
        self, array_path: str, batch_size: int
    ) -> AsyncIterator[list[list[int]]]: ...
    @property
    def store(self) -> PyStore: ...
    @property
    def config(self) -> RepositoryConfig: ...
    def merge(self, other: PySession) -> None: ...
    async def merge_async(self, other: PySession) -> None: ...
    def commit(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> str: ...
    async def commit_async(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> str: ...
    def flush(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
    ) -> str: ...
    async def flush_async(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
    ) -> str: ...
    def rebase(self, solver: ConflictSolver) -> None: ...
    async def rebase_async(self, solver: ConflictSolver) -> None: ...

class PyStore:
    @classmethod
    def from_bytes(cls, data: bytes) -> PyStore: ...
    def __eq__(self, value: object) -> bool: ...
    @property
    def read_only(self) -> bool: ...
    @property
    def session(self) -> PySession: ...
    def as_bytes(self) -> bytes: ...
    async def is_empty(self, prefix: str) -> bool: ...
    async def clear(self) -> None: ...
    def sync_clear(self) -> None: ...
    async def get(
        self, key: str, byte_range: tuple[int | None, int | None] | None = None
    ) -> bytes: ...
    async def get_partial_values(
        self, key_ranges: list[tuple[str, tuple[int | None, int | None]]]
    ) -> list[bytes]: ...
    async def exists(self, key: str) -> bool: ...
    @property
    def supports_writes(self) -> bool: ...
    @property
    def supports_consolidated_metadata(self) -> bool: ...
    @property
    def supports_deletes(self) -> bool: ...
    async def set(self, key: str, value: bytes) -> None: ...
    async def set_if_not_exists(self, key: str, value: bytes) -> None: ...
    def set_virtual_ref(
        self,
        key: str,
        location: str,
        offset: int,
        length: int,
        checksum: str | datetime.datetime | None,
        validate_container: bool,
    ) -> None: ...
    async def set_virtual_ref_async(
        self,
        key: str,
        location: str,
        offset: int,
        length: int,
        checksum: str | datetime.datetime | None,
        validate_container: bool = False,
    ) -> None: ...
    def set_virtual_refs(
        self,
        array_path: str,
        chunks: list[VirtualChunkSpec],
        validate_containers: bool,
    ) -> list[tuple[int, ...]] | None: ...
    async def set_virtual_refs_async(
        self,
        array_path: str,
        chunks: list[VirtualChunkSpec],
        validate_containers: bool,
    ) -> list[tuple[int, ...]] | None: ...
    async def delete(self, key: str) -> None: ...
    async def delete_dir(self, prefix: str) -> None: ...
    @property
    def supports_partial_writes(self) -> bool: ...
    async def set_partial_values(
        self, key_start_values: list[tuple[str, int, bytes]]
    ) -> None: ...
    @property
    def supports_listing(self) -> bool: ...
    def list(self) -> PyAsyncStringGenerator: ...
    def list_prefix(self, prefix: str) -> PyAsyncStringGenerator: ...
    def list_dir(self, prefix: str) -> PyAsyncStringGenerator: ...
    async def getsize(self, key: str) -> int: ...
    async def getsize_prefix(self, prefix: str) -> int: ...

class PyAsyncStringGenerator(AsyncGenerator[str, None], metaclass=abc.ABCMeta):
    def __aiter__(self) -> PyAsyncStringGenerator: ...
    async def __anext__(self) -> str: ...

class ManifestFileInfo:
    """Manifest file metadata"""

    @property
    def id(self) -> str:
        """The manifest id"""
        ...
    @property
    def size_bytes(self) -> int:
        """The size in bytes of the"""
        ...
    @property
    def num_chunk_refs(self) -> int:
        """The number of chunk references contained in this manifest"""
        ...

class SnapshotInfo:
    """Metadata for a snapshot"""
    @property
    def id(self) -> str:
        """The snapshot ID"""
        ...
    @property
    def parent_id(self) -> str | None:
        """The snapshot ID"""
        ...
    @property
    def written_at(self) -> datetime.datetime:
        """
        The timestamp when the snapshot was written
        """
        ...
    @property
    def message(self) -> str:
        """
        The commit message of the snapshot
        """
        ...
    @property
    def metadata(self) -> dict[str, Any]:
        """
        The metadata of the snapshot
        """
        ...
    @property
    def manifests(self) -> list[ManifestFileInfo]:
        """
        The manifests linked to this snapshot
        """
        ...

class PyAsyncSnapshotGenerator(AsyncGenerator[SnapshotInfo, None], metaclass=abc.ABCMeta):
    def __aiter__(self) -> PyAsyncSnapshotGenerator: ...
    async def __anext__(self) -> SnapshotInfo: ...

class S3StaticCredentials:
    """Credentials for an S3 storage backend

    Attributes:
        access_key_id: str
            The access key ID to use for authentication.
        secret_access_key: str
            The secret access key to use for authentication.
        session_token: str | None
            The session token to use for authentication.
        expires_after: datetime.datetime | None
            Optional, the expiration time of the credentials.
    """

    access_key_id: str
    secret_access_key: str
    session_token: str | None
    expires_after: datetime.datetime | None

    def __init__(
        self,
        access_key_id: str,
        secret_access_key: str,
        session_token: str | None = None,
        expires_after: datetime.datetime | None = None,
    ):
        """
        Create a new `S3StaticCredentials` object

        Parameters
        ----------
        access_key_id: str
            The access key ID to use for authentication.
        secret_access_key: str
            The secret access key to use for authentication.
        session_token: str | None
            Optional, the session token to use for authentication.
        expires_after: datetime.datetime | None
            Optional, the expiration time of the credentials.
        """
        ...

class S3Credentials:
    """Credentials for an S3 storage backend"""
    class FromEnv:
        """Uses credentials from environment variables"""
        def __init__(self) -> None: ...

    class Anonymous:
        """Does not sign requests, useful for public buckets"""
        def __init__(self) -> None: ...

    class Static:
        """Uses s3 credentials without expiration

        Parameters
        ----------
        credentials: S3StaticCredentials
            The credentials to use for authentication.
        """
        def __init__(self, credentials: S3StaticCredentials) -> None: ...

    class Refreshable:
        """Allows for an outside authority to pass in a function that can be used to provide credentials.

        This is useful for credentials that have an expiration time, or are otherwise not known ahead of time.

        Parameters
        ----------
        pickled_function: bytes
            The pickled function to use to provide credentials.
        current: S3StaticCredentials
            The initial credentials. They will be returned the first time credentials
            are requested and then deleted.
        """
        def __init__(
            self, pickled_function: bytes, current: S3StaticCredentials | None = None
        ) -> None: ...

AnyS3Credential = (
    S3Credentials.Static
    | S3Credentials.Anonymous
    | S3Credentials.FromEnv
    | S3Credentials.Refreshable
)

class GcsBearerCredential:
    """Credentials for a google cloud storage backend

    This is a bearer token that has an expiration time.
    """

    def __init__(
        self, bearer: str, *, expires_after: datetime.datetime | None = None
    ) -> None:
        """Create a GcsBearerCredential object

        Parameters
        ----------
        bearer: str
            The bearer token to use for authentication.
        expires_after: datetime.datetime | None
            The expiration time of the bearer token.
        """

    @property
    def bearer(self) -> str: ...
    @property
    def expires_after(self) -> datetime.datetime | None: ...

class GcsStaticCredentials:
    """Credentials for a google cloud storage backend"""
    class ServiceAccount:
        """Credentials for a google cloud storage backend using a service account json file

        Parameters
        ----------
        path: str
            The path to the service account json file.
        """
        def __init__(self, path: str) -> None: ...

    class ServiceAccountKey:
        """Credentials for a google cloud storage backend using a a serialized service account key

        Parameters
        ----------
        key: str
            The serialized service account key.
        """
        def __init__(self, key: str) -> None: ...

    class ApplicationCredentials:
        """Credentials for a google cloud storage backend using application default credentials

        Parameters
        ----------
        path: str
            The path to the application default credentials (ADC) file.
        """
        def __init__(self, path: str) -> None: ...

    class BearerToken:
        """Credentials for a google cloud storage backend using a bearer token

        Parameters
        ----------
        token: str
            The bearer token to use for authentication.
        """
        def __init__(self, token: str) -> None: ...

AnyGcsStaticCredential = (
    GcsStaticCredentials.ServiceAccount
    | GcsStaticCredentials.ServiceAccountKey
    | GcsStaticCredentials.ApplicationCredentials
    | GcsStaticCredentials.BearerToken
)

class GcsCredentials:
    """Credentials for a google cloud storage backend

    This can be used to authenticate with a google cloud storage backend.
    """
    class Anonymous:
        """Uses anonymous credentials"""
        def __init__(self) -> None: ...

    class FromEnv:
        """Uses credentials from environment variables"""
        def __init__(self) -> None: ...

    class Static:
        """Uses gcs credentials without expiration"""
        def __init__(self, credentials: AnyGcsStaticCredential) -> None: ...

    class Refreshable:
        """Allows for an outside authority to pass in a function that can be used to provide credentials.

        This is useful for credentials that have an expiration time, or are otherwise not known ahead of time.
        """
        def __init__(
            self, pickled_function: bytes, current: GcsBearerCredential | None = None
        ) -> None: ...

AnyGcsCredential = (
    GcsCredentials.Anonymous
    | GcsCredentials.FromEnv
    | GcsCredentials.Static
    | GcsCredentials.Refreshable
)

class AzureStaticCredentials:
    """Credentials for an azure storage backend"""
    class AccessKey:
        """Credentials for an azure storage backend using an access key

        Parameters
        ----------
        key: str
            The access key to use for authentication.
        """
        def __init__(self, key: str) -> None: ...

    class SasToken:
        """Credentials for an azure storage backend using a shared access signature token

        Parameters
        ----------
        token: str
            The shared access signature token to use for authentication.
        """
        def __init__(self, token: str) -> None: ...

    class BearerToken:
        """Credentials for an azure storage backend using a bearer token

        Parameters
        ----------
        token: str
            The bearer token to use for authentication.
        """
        def __init__(self, token: str) -> None: ...

AnyAzureStaticCredential = (
    AzureStaticCredentials.AccessKey
    | AzureStaticCredentials.SasToken
    | AzureStaticCredentials.BearerToken
)

class AzureCredentials:
    """Credentials for an azure storage backend

    This can be used to authenticate with an azure storage backend.
    """
    class FromEnv:
        """Uses credentials from environment variables"""
        def __init__(self) -> None: ...

    class Static:
        """Uses azure credentials without expiration"""
        def __init__(self, credentials: AnyAzureStaticCredential) -> None: ...

AnyAzureCredential = AzureCredentials.FromEnv | AzureCredentials.Static

class Credentials:
    class S3:
        def __init__(self, credentials: AnyS3Credential) -> None: ...

    class Gcs:
        def __init__(self, credentials: GcsCredentials) -> None: ...

    class Azure:
        def __init__(self, credentials: AzureCredentials) -> None: ...

AnyCredential = Credentials.S3 | Credentials.Gcs | Credentials.Azure

class Storage:
    """Storage configuration for an IcechunkStore

    Currently supports memory, filesystem S3, azure blob, and google cloud storage backends.
    Use the following methods to create a Storage object with the desired backend.

    Ex:
    ```
    storage = icechunk.in_memory_storage()
    storage = icechunk.local_filesystem_storage("/path/to/root")
    storage = icechunk.s3_storage("bucket", "prefix", ...)
    storage = icechunk.gcs_storage("bucket", "prefix", ...)
    storage = icechunk.azure_storage("container", "prefix", ...)
    ```
    """

    @classmethod
    def new_s3(
        cls,
        config: S3Options,
        bucket: str,
        prefix: str | None,
        credentials: AnyS3Credential | None = None,
    ) -> Storage: ...
    @classmethod
    def new_s3_object_store(
        cls,
        config: S3Options,
        bucket: str,
        prefix: str | None,
        credentials: AnyS3Credential | None = None,
    ) -> Storage: ...
    @classmethod
    def new_tigris(
        cls,
        config: S3Options,
        bucket: str,
        prefix: str | None,
        use_weak_consistency: bool,
        credentials: AnyS3Credential | None = None,
    ) -> Storage: ...
    @classmethod
    def new_in_memory(cls) -> Storage: ...
    @classmethod
    def new_local_filesystem(cls, path: str) -> Storage: ...
    @classmethod
    def new_gcs(
        cls,
        bucket: str,
        prefix: str | None,
        credentials: AnyGcsCredential | None = None,
        *,
        config: dict[str, str] | None = None,
    ) -> Storage: ...
    @classmethod
    def new_r2(
        cls,
        bucket: str | None,
        prefix: str | None,
        account_id: str | None,
        credentials: AnyS3Credential | None = None,
        *,
        config: S3Options,
    ) -> Storage: ...
    @classmethod
    def new_azure_blob(
        cls,
        account: str,
        container: str,
        prefix: str,
        credentials: AnyAzureCredential | None = None,
        *,
        config: dict[str, str] | None = None,
    ) -> Storage: ...
    def __repr__(self) -> str: ...
    def default_settings(self) -> StorageSettings: ...

class VersionSelection(Enum):
    """Enum for selecting the which version of a conflict

    Attributes
    ----------
    Fail: int
        Fail the rebase operation
    UseOurs: int
        Use the version from the source store
    UseTheirs: int
        Use the version from the target store
    """

    Fail = 0
    UseOurs = 1
    UseTheirs = 2

class ConflictSolver:
    """An abstract conflict solver that can be used to detect or resolve conflicts between two stores

    This should never be used directly, but should be subclassed to provide specific conflict resolution behavior
    """

    ...

class BasicConflictSolver(ConflictSolver):
    """A basic conflict solver that allows for simple configuration of resolution behavior

    This conflict solver allows for simple configuration of resolution behavior for conflicts that may occur during a rebase operation.
    It will attempt to resolve a limited set of conflicts based on the configuration options provided.

    - When a chunk conflict is encountered, the behavior is determined by the `on_chunk_conflict` option
    - When an array is deleted that has been updated, `fail_on_delete_of_updated_array` will determine whether to fail the rebase operation
    - When a group is deleted that has been updated, `fail_on_delete_of_updated_group` will determine whether to fail the rebase operation
    """

    def __init__(
        self,
        *,
        on_chunk_conflict: VersionSelection = VersionSelection.UseOurs,
        fail_on_delete_of_updated_array: bool = False,
        fail_on_delete_of_updated_group: bool = False,
    ) -> None:
        """Create a BasicConflictSolver object with the given configuration options

        Parameters
        ----------
        on_chunk_conflict: VersionSelection
            The behavior to use when a chunk conflict is encountered, by default VersionSelection.use_theirs()
        fail_on_delete_of_updated_array: bool
            Whether to fail when a chunk is deleted that has been updated, by default False
        fail_on_delete_of_updated_group: bool
            Whether to fail when a group is deleted that has been updated, by default False
        """
        ...

class ConflictDetector(ConflictSolver):
    """A conflict solver that can be used to detect conflicts between two stores, but does not resolve them

    Where the `BasicConflictSolver` will attempt to resolve conflicts, the `ConflictDetector` will only detect them. This means
    that during a rebase operation the `ConflictDetector` will raise a `RebaseFailed` error if any conflicts are detected, and
    allow the rebase operation to be retried with a different conflict resolution strategy. Otherwise, if no conflicts are detected
    the rebase operation will succeed.
    """

    def __init__(self) -> None: ...

class IcechunkError(Exception):
    """Base class for all Icechunk errors"""

    @property
    def message(self) -> str: ...

class ConflictError(Exception):
    """An error that occurs when a conflict is detected"""

    @property
    def expected_parent(self) -> str:
        """The expected parent snapshot ID.

        This is the snapshot ID that the session was based on when the
        commit operation was called.
        """
        ...
    @property
    def actual_parent(self) -> str:
        """
        The actual parent snapshot ID of the branch that the session attempted to commit to.

        When the session is based on a branch, this is the snapshot ID of the branch tip. If this
        error is raised, it means the branch was modified and committed by another session after
        the session was created.
        """
        ...
    ...

__version__: str

class ConflictType(Enum):
    """Type of conflict detected"""

    NewNodeConflictsWithExistingNode = (1,)
    """A new node conflicts with an existing node"""

    NewNodeInInvalidGroup = (2,)
    """A new node is in an invalid group"""

    ZarrMetadataDoubleUpdate = (3,)
    """A zarr metadata update conflicts with an existing zarr metadata update"""

    ZarrMetadataUpdateOfDeletedArray = (4,)
    """A zarr metadata update is attempted on a deleted array"""

    ZarrMetadataUpdateOfDeletedGroup = (5,)
    """A zarr metadata update is attempted on a deleted group"""

    ChunkDoubleUpdate = (6,)
    """A chunk update conflicts with an existing chunk update"""

    ChunksUpdatedInDeletedArray = (7,)
    """Chunks are updated in a deleted array"""

    ChunksUpdatedInUpdatedArray = (8,)
    """Chunks are updated in an updated array"""

    DeleteOfUpdatedArray = (9,)
    """A delete is attempted on an updated array"""

    DeleteOfUpdatedGroup = (10,)
    """A delete is attempted on an updated group"""

class Conflict:
    """A conflict detected between snapshots"""

    @property
    def conflict_type(self) -> ConflictType:
        """The type of conflict detected

        Returns:
            ConflictType: The type of conflict detected
        """
        ...

    @property
    def path(self) -> str:
        """The path of the node that caused the conflict

        Returns:
            str: The path of the node that caused the conflict
        """
        ...

    @property
    def conflicted_chunks(self) -> list[list[int]] | None:
        """If the conflict is a chunk conflict, this will return the list of chunk indices that are in conflict

        Returns:
            list[list[int]] | None: The list of chunk indices that are in conflict
        """
        ...

class RebaseFailedError(IcechunkError):
    """An error that occurs when a rebase operation fails"""

    @property
    def snapshot(self) -> str:
        """The snapshot ID that the session was rebased to"""
        ...

    @property
    def conflicts(self) -> list[Conflict]:
        """The conflicts that occurred during the rebase operation

        Returns:
            list[Conflict]: The conflicts that occurred during the rebase operation
        """
    ...

def initialize_logs() -> None:
    """
    Initialize the logging system for the library.

    Reads the value of the environment variable ICECHUNK_LOG to obtain the filters.
    This is autamtically called on `import icechunk`.
    """
    ...

def set_logs_filter(log_filter_directive: str | None) -> None:
    """
    Set filters and log levels for the different modules.

    Examples:
      - set_logs_filter("trace")  # trace level for all modules
      - set_logs_filter("error")  # error level for all modules
      - set_logs_filter("icechunk=debug,info")  # debug level for icechunk, info for everything else

    Full spec for the log_filter_directive syntax is documented in
    https://docs.rs/tracing-subscriber/latest/tracing_subscriber/filter/struct.EnvFilter.html#directives

    Parameters
    ----------
    log_filter_directive: str | None
        The comma separated list of directives for modules and log levels.
        If None, the directive will be read from the environment variable
        ICECHUNK_LOG
    """
    ...

def spec_version() -> int:
    """
    The version of the Icechunk specification that the library is compatible with.

    Returns:
        int: The version of the Icechunk specification that the library is compatible with
    """
    ...



import time
from datetime import UTC, datetime
from typing import cast

import pytest

import icechunk as ic
import zarr
from tests.conftest import get_minio_client


def mk_repo() -> tuple[str, ic.Repository]:
    prefix = "test-repo__" + str(time.time())
    repo = ic.Repository.create(
        storage=ic.s3_storage(
            endpoint_url="http://localhost:9000",
            allow_http=True,
            force_path_style=True,
            region="us-east-1",
            bucket="testbucket",
            prefix=prefix,
            access_key_id="minio123",
            secret_access_key="minio123",
        ),
        config=ic.RepositoryConfig(inline_chunk_threshold_bytes=0),
    )
    return (prefix, repo)


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow")
@pytest.mark.parametrize("use_async", [True, False])
async def test_expire_and_gc(use_async: bool) -> None:
    prefix, repo = mk_repo()

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    array = group.create_array(
        "array",
        shape=(1000),
        chunks=(10),
        dtype="i4",
        fill_value=-1,
    )
    session.commit("array created")

    for i in range(20):
        session = repo.writable_session("main")
        store = session.store
        group = zarr.open_group(store=store)
        array = cast(zarr.core.array.Array, group["array"])
        array[i] = i
        session.commit(f"written coord {i}")

    old = datetime.now(UTC)

    session = repo.writable_session("main")
    store = session.store
    group = zarr.open_group(store=store)
    array = cast(zarr.core.array.Array, group["array"])
    array[999] = 0
    session.commit("written coord 999")

    client = get_minio_client()

    # repo initial snap + array creation + 20 old version + 1 new version
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/snapshots")[
                "Contents"
            ]
        )
        == 23
    )
    # 20 old chunks + 1 new
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/chunks")[
                "Contents"
            ]
        )
        == 21
    )
    # 21 commits that modify chunks
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/manifests")[
                "Contents"
            ]
        )
        == 21
    )
    # 22 commits total
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/transactions")[
                "Contents"
            ]
        )
        == 22
    )

    if use_async:
        expired_snapshots = await repo.expire_snapshots_async(old)
    else:
        expired_snapshots = repo.expire_snapshots(old)
    # empty array + 20 old versions
    assert len(expired_snapshots) == 21

    space_before = 0
    for obj in client.list_objects(Bucket="testbucket", Prefix=f"{prefix}")["Contents"]:
        space_before += obj["Size"]

    # let's run GC using dry_run = True
    if use_async:
        gc_result = await repo.garbage_collect_async(old, dry_run=True)
    else:
        gc_result = repo.garbage_collect(old, dry_run=True)
    space_after = 0
    for obj in client.list_objects(Bucket="testbucket", Prefix=f"{prefix}")["Contents"]:
        space_after += obj["Size"]

    assert space_before == space_after
    # there were 21 chunks, and we need 3 alive (for indexes 0..20 and 999)
    assert gc_result.chunks_deleted == 18
    # there were 23 snapshots, we need the initial one and the latest version only
    assert gc_result.snapshots_deleted == 21
    # there were 21 manifests, we need only 1
    assert gc_result.manifests_deleted == 20
    # not implemented yet
    assert gc_result.attributes_deleted == 0
    # same number as snapshots
    assert gc_result.transaction_logs_deleted == 21

    # now let's run real GC, no dry_run.
    if use_async:
        gc_result = await repo.garbage_collect_async(old)
    else:
        gc_result = repo.garbage_collect(old)

    space_after = 0
    for obj in client.list_objects(Bucket="testbucket", Prefix=f"{prefix}")["Contents"]:
        space_after += obj["Size"]

    assert space_before - gc_result.bytes_deleted == space_after
    # there were 21 chunks, and we need 3 alive (for indexes 0..20 and 999)
    assert gc_result.chunks_deleted == 18
    # there were 23 snapshots, we need the initial one and the latest version only
    assert gc_result.snapshots_deleted == 21
    # there were 21 manifests, we need only 1
    assert gc_result.manifests_deleted == 20
    # not implemented yet
    assert gc_result.attributes_deleted == 0
    # same number as snapshots
    assert gc_result.transaction_logs_deleted == 21

    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/snapshots")[
                "Contents"
            ]
        )
        == 2
    )
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/chunks")[
                "Contents"
            ]
        )
        == 3
    )
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/manifests")[
                "Contents"
            ]
        )
        == 1
    )
    assert (
        len(
            client.list_objects(Bucket="testbucket", Prefix=f"{prefix}/transactions")[
                "Contents"
            ]
        )
        == 1
    )

    # we can still read the array
    session = repo.readonly_session(branch="main")
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    array = cast(zarr.core.array.Array, group["array"])
    assert array[999] == 0
    for i in range(20):
        assert array[i] == i



from __future__ import annotations

import pickle
from pathlib import Path
from typing import Any, TypeVar

import pytest

from icechunk import IcechunkError, IcechunkStore, local_filesystem_storage
from icechunk.repository import Repository
from zarr.abc.store import OffsetByteRequest, RangeByteRequest, Store, SuffixByteRequest
from zarr.core.buffer import Buffer, cpu, default_buffer_prototype
from zarr.core.sync import _collect_aiterator, collect_aiterator
from zarr.testing.store import StoreTests
from zarr.testing.utils import assert_bytes_equal

DEFAULT_GROUP_METADATA = b'{"zarr_format":3,"node_type":"group"}'
ARRAY_METADATA = (
    b'{"zarr_format":3,"node_type":"array","attributes":{"foo":42},'
    b'"shape":[2,2,2],"data_type":"int32","chunk_grid":{"name":"regular","configuration":{"chunk_shape":[1,1,1]}},'
    b'"chunk_key_encoding":{"name":"default","configuration":{"separator":"/"}},"fill_value":0,'
    b'"codecs":[{"name":"mycodec","configuration":{"foo":42}}],"storage_transformers":[{"name":"mytransformer","configuration":{"bar":43}}],"dimension_names":["x","y","t"]}'
)

S = TypeVar("S", bound=Store)


class TestIcechunkStore(StoreTests[IcechunkStore, cpu.Buffer]):
    store_cls = IcechunkStore
    buffer_cls = cpu.Buffer

    async def set(self, store: IcechunkStore, key: str, value: Buffer) -> None:
        await store._store.set(key, value.to_bytes())

    async def get(self, store: IcechunkStore, key: str) -> Buffer | None:
        try:
            result = await store._store.get(key)
            if result is None:
                return None
        except ValueError as _e:
            # Zarr python expects None to be returned if the key does not exist
            # but an IcechunkStore returns an error if the key does not exist
            return None

        return self.buffer_cls.from_bytes(result)

    @pytest.fixture
    def store_kwargs(self, tmpdir: Path) -> dict[str, Any]:
        kwargs = {
            "storage": local_filesystem_storage(f"{tmpdir}/store_test"),
            "read_only": False,
        }
        return kwargs

    @pytest.fixture
    async def store(self, store_kwargs: dict[str, Any]) -> IcechunkStore:
        read_only = store_kwargs.pop("read_only")
        repo = Repository.open_or_create(**store_kwargs)
        if read_only:
            session = repo.readonly_session(branch="main")
        else:
            session = repo.writable_session("main")
        return session.store

    @pytest.fixture
    async def store_not_open(self, store_kwargs: dict[str, Any]) -> IcechunkStore:
        read_only = store_kwargs.pop("read_only")
        repo = Repository.open_or_create(**store_kwargs)
        if read_only:
            session = repo.readonly_session(branch="main")
        else:
            session = repo.writable_session("main")

        store = session.store
        store._is_open = False
        return store

    def test_store_eq(self, store: IcechunkStore, store_kwargs: dict[str, Any]) -> None:
        # check self equality
        assert store == store

        # check store equality with same inputs
        # asserting this is important for being able to compare (de)serialized stores
        repo = Repository.open(**store_kwargs)
        if store.read_only:
            session = repo.readonly_session(branch="main")
        else:
            session = repo.writable_session("main")
        store2 = session.store
        # stores dont point to the same session instance, so they are not equal
        assert store != store2

    @pytest.mark.skip(reason="Not implemented")
    def test_store_repr(self, store: IcechunkStore) -> None:
        pass

    @pytest.mark.parametrize("read_only", [True, False])
    async def test_store_open_read_only(
        self, store: IcechunkStore, store_kwargs: dict[str, Any], read_only: bool
    ) -> None:
        repo = Repository.open(**store_kwargs)
        if read_only:
            session = repo.readonly_session(branch="main")
        else:
            session = repo.writable_session("main")
        store = session.store
        assert store._is_open
        assert store.read_only == read_only

    async def test_read_only_store_raises(
        self, store: IcechunkStore, store_kwargs: dict[str, Any]
    ) -> None:
        kwargs = {**store_kwargs}
        repo = Repository.open(**kwargs)
        session = repo.readonly_session(branch="main")
        store = session.store

        assert store.read_only

        # set
        with pytest.raises(IcechunkError):
            await store.set("foo", self.buffer_cls.from_bytes(b"bar"))

        # delete
        with pytest.raises(IcechunkError):
            await store.delete("foo")

    @pytest.mark.skip(
        reason="icechunk requires opting-in to pickling at the session level"
    )
    def test_serializable_store(self, store: IcechunkStore) -> None:
        foo = pickle.dumps(store)
        loaded = pickle.loads(foo)
        # pickled stores dont point to the same session instance, so they are not equal
        assert loaded != store

    @pytest.mark.skip(
        reason="icechunk read-only follows the read-only flag of the session"
    )
    async def test_with_read_only_store(self, open_kwargs: dict[str, Any]) -> None:
        pass

    async def test_get_not_open(self, store_not_open: IcechunkStore) -> None:
        """
        Ensure that data can be read from the store that isn't yet open using the store.get method.
        """
        assert not store_not_open._is_open

        # create an array
        await store_not_open.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        data_buf = self.buffer_cls.from_bytes(b"\x01\x02\x03\x04")
        key = "c/0/0/0"
        await self.set(store_not_open, key, data_buf)
        observed = await store_not_open.get(key, prototype=default_buffer_prototype())
        assert_bytes_equal(observed, data_buf)

    async def test_get_raises(self, store: IcechunkStore) -> None:
        """
        Ensure that a ValueError is raise for invalid byte range syntax
        """
        # create an array
        await store.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        data_buf = self.buffer_cls.from_bytes(b"\x01\x02\x03\x04")
        await self.set(store, "c/0/0/0", data_buf)
        with pytest.raises(
            (ValueError, TypeError), match=r"Unexpected byte_range, got.*"
        ):
            await store.get(
                "c/0/0/0", prototype=default_buffer_prototype(), byte_range=(0, 2)
            )  # type: ignore[arg-type]

    @pytest.mark.xfail(reason="Not implemented")
    async def test_store_context_manager(self, open_kwargs: dict[str, Any]) -> None:
        # Test that the context manager closes the store
        with await self.store_cls.open(**open_kwargs) as store:
            assert store._is_open
            # Test trying to open an already open store
            with pytest.raises(ValueError, match="store is already open"):
                await store._open()
        assert not store._is_open

    async def test_set_many(self, store: IcechunkStore) -> None:
        """
        Test that a dict of key : value pairs can be inserted into the store via the
        `_set_many` method.
        """
        # This test won't work without initializing the array first
        await store.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        keys = [
            "zarr.json",
            "c/0/0/0",
            # icechunk does not allow v2 keys
            # "foo/c/0.0",
            # "foo/0/0"
        ]
        # icechunk strictly checks metadata?
        data_buf = [
            self.buffer_cls.from_bytes(k.encode() if k != "zarr.json" else ARRAY_METADATA)
            for k in keys
        ]
        store_dict = dict(zip(keys, data_buf, strict=True))
        await store._set_many(store_dict.items())
        for k, v in store_dict.items():
            assert (await self.get(store, k)).to_bytes() == v.to_bytes()

    async def test_set_not_open(self, store_not_open: IcechunkStore) -> None:
        """
        Ensure that data can be written to the store that's not yet open using the store.set method.
        """
        assert not store_not_open._is_open

        # create an array
        await store_not_open.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        data_buf = self.buffer_cls.from_bytes(b"\x01\x02\x03\x04")
        key = "c/0/0/0"
        await store_not_open.set(key, data_buf)
        observed = await self.get(store_not_open, key)
        assert_bytes_equal(observed, data_buf)

    async def test_set_invalid_buffer(self, store: S) -> None:
        """
        Ensure that set raises a Type or Value Error for invalid buffer arguments.
        """
        # create an array
        await store.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        with pytest.raises(
            (ValueError, TypeError),
            match=r"\S+\.set\(\): `value` must be a Buffer instance. Got an instance of <class 'int'> instead.",
        ):
            await store.set("c/0/0/0", 0)  # type: ignore[arg-type]

    def test_store_supports_deletes(self, store: IcechunkStore) -> None:
        assert store.supports_deletes

    def test_store_supports_writes(self, store: IcechunkStore) -> None:
        assert store.supports_writes

    def test_store_supports_listing(self, store: IcechunkStore) -> None:
        assert store.supports_listing

    def test_store_supports_partial_writes(self, store: IcechunkStore) -> None:
        assert not store.supports_partial_writes

    async def test_clear(self, store: IcechunkStore) -> None:
        await self.set(
            store,
            "foo/zarr.json",
            self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA),
        )
        await store.clear()

    async def test_exists(self, store: IcechunkStore) -> None:
        # Icechunk store does not support arbitrary keys
        with pytest.raises(IcechunkError):
            await store.exists("foo")
        assert not await store.exists("foo/zarr.json")

        # Icechunk store does not support arbitrary data either
        with pytest.raises(IcechunkError):
            await store.set("foo", self.buffer_cls.from_bytes(b"bar"))

        await store.set(
            "foo/zarr.json",
            self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA),
        )
        assert await store.exists("foo/zarr.json")

    async def test_list(self, store: IcechunkStore) -> None:
        assert [k async for k in store.list()] == []
        await store.set(
            "foo/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        keys = [k async for k in store.list()]
        assert keys == ["foo/zarr.json"], keys

    async def test_list_prefix(self, store: S) -> None:
        """
        Test that the `list_prefix` method works as intended. Given a prefix, it should return
        all the keys in storage that start with this prefix.
        """
        prefixes = ("", "a/", "a/b/", "a/b/c/")
        data = self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        fname = "zarr.json"
        store_dict = {p + fname: data for p in prefixes}

        await store._set_many(store_dict.items())

        for prefix in prefixes:
            observed = tuple(sorted([_ async for _ in store.list_prefix(prefix)]))
            expected: tuple[str, ...] = ()
            for key in store_dict:
                if key.startswith(prefix):
                    expected += (key,)
            expected = tuple(sorted(expected))
            assert observed == expected

    async def test_list_empty_path(self, store: S) -> None:
        """
        Verify that list and list_prefix work correctly when path is an empty string,
        i.e. no unwanted replacement occurs.
        """
        await store.set("foo/bar/zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))
        data = self.buffer_cls.from_bytes(b"")
        store_dict = {
            "foo/bar/c/1/0/0": data,
            "foo/bar/c/0/0/0": data,
        }
        await store._set_many(store_dict.items())

        all_keys = sorted(list(store_dict.keys()) + ["foo/bar/zarr.json"])

        # Test list()
        observed_list = await _collect_aiterator(store.list())
        observed_list_sorted = sorted(observed_list)
        expected_list_sorted = all_keys
        assert observed_list_sorted == expected_list_sorted

        # Test list_prefix() with an empty prefix
        observed_prefix_empty = await _collect_aiterator(store.list_prefix(""))
        observed_prefix_empty_sorted = sorted(observed_prefix_empty)
        expected_prefix_empty_sorted = all_keys
        assert observed_prefix_empty_sorted == expected_prefix_empty_sorted

        # Test list_prefix() with a non-empty prefix
        observed_prefix = await _collect_aiterator(store.list_prefix("foo/bar/"))
        observed_prefix_sorted = sorted(observed_prefix)
        expected_prefix_sorted = sorted(k for k in all_keys if k.startswith("foo/bar/"))
        assert observed_prefix_sorted == expected_prefix_sorted

    async def test_list_dir(self, store: IcechunkStore) -> None:
        out = [k async for k in store.list_dir("")]
        assert out == []

        await store.set("zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA))

        await store.set(
            "foo/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        await store.set(
            "goo/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )

        keys_expected = ["foo", "goo", "zarr.json"]
        keys_observed = [k async for k in store.list_dir("")]
        assert set(keys_observed) == set(keys_expected)

        keys_expected = ["zarr.json"]
        keys_observed = [k async for k in store.list_dir("foo")]
        assert set(keys_observed) == set(keys_expected)

    async def test_delete(self, store: IcechunkStore) -> None:
        await store.set(
            "foo/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        assert await store.exists("foo/zarr.json")
        await store.delete("foo/zarr.json")
        assert not await store.exists("foo/zarr.json")

    async def test_delete_nonexistent_key_does_not_raise(
        self, store: IcechunkStore
    ) -> None:
        if not store.supports_deletes:
            pytest.skip("store does not support deletes")
        await store.delete("zarr.json")

    async def test_get_partial_values(
        self,
        store: IcechunkStore,
    ) -> None:
        await self.set(
            store, "zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        # read back just part of it
        values = await store.get_partial_values(
            default_buffer_prototype(),
            [
                ("zarr.json", RangeByteRequest(0, 5)),
                ("zarr.json", SuffixByteRequest(5)),
                ("zarr.json", OffsetByteRequest(10)),
            ],
        )

        assert len(values) == 3
        data = values[0].to_bytes()
        assert len(data) == 5
        assert data == DEFAULT_GROUP_METADATA[:5]

        data = values[1].to_bytes()
        assert len(data) == len(DEFAULT_GROUP_METADATA) - 5
        assert data == DEFAULT_GROUP_METADATA[:-5]

        data = values[2].to_bytes()
        assert len(data) == len(DEFAULT_GROUP_METADATA) - 10
        assert data == DEFAULT_GROUP_METADATA[10:]

    async def test_set(self, store: IcechunkStore) -> None:
        await store.set("zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA))
        assert await store.exists("zarr.json")
        result = await self.get(store, "zarr.json")
        assert result.to_bytes() == DEFAULT_GROUP_METADATA

    async def test_get(self, store: IcechunkStore) -> None:
        await self.set(
            store, "zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        assert await store.exists("zarr.json")
        result = await store.get("zarr.json", default_buffer_prototype())
        assert result is not None
        assert result.to_bytes() == DEFAULT_GROUP_METADATA

    async def test_get_many(self, store: IcechunkStore) -> None:
        """
        Ensure that multiple keys can be retrieved at once with the _get_many method.
        """
        await store.set("zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        keys = [
            "c/0/0/0",
            "c/0/0/1",
            "c/0/1/0",
            "c/0/1/1",
            "c/1/0/0",
            "c/1/0/1",
            "c/1/1/0",
            "c/1/1/1",
        ]
        values = [bytes(i) for i, _ in enumerate(keys)]
        for k, v in zip(keys, values, strict=False):
            await self.set(store, k, self.buffer_cls.from_bytes(v))
        observed_buffers = collect_aiterator(
            store._get_many(
                zip(
                    keys,
                    (default_buffer_prototype(),) * len(keys),
                    (None,) * len(keys),
                    strict=False,
                )
            )
        )
        observed_kvs = sorted(((k, b.to_bytes()) for k, b in observed_buffers))  # type: ignore[union-attr]
        expected_kvs = sorted(((k, b) for k, b in zip(keys, values, strict=False)))
        assert observed_kvs == expected_kvs

    async def test_set_if_not_exists(self, store: IcechunkStore) -> None:
        key = "zarr.json"
        data_buf = self.buffer_cls.from_bytes(ARRAY_METADATA)
        await self.set(store, key, data_buf)

        new = self.buffer_cls.from_bytes(b"1111")

        # no error even though the data is invalid and the metadata exists
        await store.set_if_not_exists(key, new)

        result = await store.get(key, default_buffer_prototype())
        assert result == data_buf

        await store.set_if_not_exists("c/0/0/0", new)  # no error

        result = await store.get("c/0/0/0", default_buffer_prototype())
        assert result == new

    async def test_is_empty(self, store: IcechunkStore) -> None:
        assert await store.is_empty("")
        await self.set(
            store, "zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        await self.set(
            store, "foo/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        await self.set(
            store, "foo/bar/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        assert not await store.is_empty("")
        assert await store.is_empty("fo")  # codespell:ignore
        assert not await store.is_empty("foo/")
        assert not await store.is_empty("foo")
        assert await store.is_empty("spam/")

    async def test_delete_dir(self, store: IcechunkStore) -> None:
        await store.set("zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA))
        await store.set(
            "foo-bar/zarr.json", self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        )
        await store.set("foo/zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))
        await store.set("foo/c/0/0/0", self.buffer_cls.from_bytes(b"chun"))
        await store.delete_dir("foo")
        assert await store.exists("zarr.json")
        assert await store.exists("foo-bar/zarr.json")
        assert not await store.exists("foo/zarr.json")
        assert not await store.exists("foo/c/0/0/0")

    async def test_getsize(self, store: IcechunkStore) -> None:
        key = "k/zarr.json"
        data = self.buffer_cls.from_bytes(DEFAULT_GROUP_METADATA)
        await self.set(store, key, data)

        result = await store.getsize(key)
        assert isinstance(result, int)
        assert result == len(DEFAULT_GROUP_METADATA)

    async def test_getsize_prefix(self, store: IcechunkStore) -> None:
        prefix = "array"
        await store.set(f"{prefix}/zarr.json", self.buffer_cls.from_bytes(ARRAY_METADATA))

        keys = [
            f"{prefix}/c/0/0/0",
            f"{prefix}/c/0/0/1",
            f"{prefix}/c/0/1/0",
            f"{prefix}/c/0/1/1",
            f"{prefix}/c/1/0/0",
            f"{prefix}/c/1/0/1",
            f"{prefix}/c/1/1/0",
            f"{prefix}/c/1/1/1",
        ]
        values = [bytes(i) for i, _ in enumerate(keys)]
        for k, v in zip(keys, values, strict=False):
            await self.set(store, k, self.buffer_cls.from_bytes(v))

        result = await store.getsize_prefix(prefix)
        assert isinstance(result, int)
        assert result == sum(len(v) for v in values) + len(ARRAY_METADATA)

    async def test_getsize_raises(self, store: IcechunkStore) -> None:
        # TODO: This maybe should be a FileNotFoundError instead of an IcechunkError
        with pytest.raises(IcechunkError):
            await store.getsize("not-a-real-key")
